#####################################
# Example of analysis of data
#####################################

import numpy as np
import pylab as plt
from boutdata.collect import collect

path = "./data/"
var = collect("P", path=path)

dcvar = np.mean(var, axis=3)
rmsvar = np.sqrt(np.mean(var**2, axis=3) - dcvar**2)

plt.figure()
plt.plot(rmsvar[:, 34, 32])
plt.show(block=False)
fvar = np.fft.rfft(var, axis=3)

plt.figure()
plt.plot(abs(fvar[:, 34, 32, 1:10]))
plt.show(block=False)


plt.figure()
plt.semilogy(abs(fvar[:, 34, 32, 1:7]))
plt.show(block=False)

plt.show()



from boutdata.collect import collect
from boututils.moment_xyzt import moment_xyzt
from boututils.datafile import DataFile
from pylab import save, figure, plot, title, xlabel, ylabel, show, tight_layout
from elm_size import elm_size

path = "./data"

t_array = collect("t_array", path=path)
save("t_array.dat", t_array)
p0 = collect("P0", path=path)
save("p0.dat", p0)


# n0=collect('n0', path=path)
# save('n0.dat', n0
# ti0=collect('ti0', path=path)
# save('ti0.dat', ti0)
# te0=collect('te0', path=path)
# save('te0.idl.dat', te0)

with DataFile("./cbm18_dens8.grid_nx68ny64.nc") as f:
    gfile = {v: f.read(v) for v in f.keys()}

p = collect("P", path=path)
save("p.dat", p)
res = moment_xyzt(p, "RMS", "DC")
rmsp = res.rms
dcp = res.dc
save("rmsp.dat", rmsp)
save("dcp.dat", dcp)
elmsp = elm_size(dcp, p0, gfile, yind=32, Bbar=gfile["bmag"])
save("elmsp.dat", elmsp)

figure(0)
plot(t_array, elmsp.s2, "k-")
xlabel("t/Ta")
ylabel("Elm size")
title("Elm size, P")
tight_layout()
show()


phi = collect("phi", path=path)
save("phi.dat", phi)
res = moment_xyzt(phi, "DC", "RMS")
save("dcphi.dat", res.dc)
save("rmsphi.dat", res.rms)



import matplotlib.pyplot as plt
from boutdata import collect
import numpy as np

run = True  # Run the simulations?
nproc = 2

# Note: Data from fci-wave-logn examples commented out.

data_noboundary = [
    ("div", "Model 1 (density, point interpolation)"),
    ("div-integrate", "Model 2 (density, area integration)"),
    ("logn", "Model 3 (log density, area integration)"),
    # ,("../fci-wave-logn/div-integrate", "Model 5 (velocity, log density, area integration)")
]

data_boundary = [
    ("boundary", "Model 2 (density, momentum)"),
    ("boundary-logn", "Model 3 (log density, momentum)"),
    # ,("../fci-wave-logn/boundary", "Model 5 (log density, velocity)")
]

# Change this to select no boundary or boundary cases
data = data_noboundary

if run:
    from boututils.run_wrapper import shell_safe, launch_safe

    shell_safe("make > make.log")

    for path, label in data:
        launch_safe("./fci-wave -d " + path, nproc=nproc, pipe=False)

# Collect the results into a dictionary
sum_n_B = {}

for path, label in data:
    n = collect("n", path=path)
    Bxyz = collect("Bxyz", path=path)

    time = collect("t_array", path=path)

    nt, nx, ny, nz = n.shape

    n_B = np.ndarray(nt)
    for t in range(nt):
        n_B[t] = np.sum(n[t, :, :, :] / Bxyz)

    sum_n_B[path] = (time, n_B)

    # Plot the density at the final time

    plt.figure()
    plt.contourf(n[-1, :, 0, :].T, 100)
    plt.colorbar()
    plt.xlabel("Major radius")
    plt.ylabel("Height")
    plt.title("Density n, " + label)
    plt.savefig(path + ".pdf")
    plt.show()

# Make a plot comparing total sum density / B

plt.figure()
for path, label in data:
    time, n_B = sum_n_B[path]
    plt.plot(time, n_B, label=label)
plt.legend()
plt.xlabel("Time")
plt.ylabel("Sum(n / B)")
plt.savefig("compare-density.pdf")

plt.show()



# MPark.Variant
#
# Copyright Michael Park, 2015-2017
#
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE.md or copy at http://boost.org/LICENSE_1_0.txt)

import os
import pprint
import subprocess

result = {}

std_flags = os.getenv('STDFLAGS')
for std_flag in std_flags.split() if std_flags is not None else ['']:
  os.environ['CXXFLAGS'] = std_flag
  for exceptions in ['OFF', 'ON']:
    for build_type in ['Debug', 'Release']:
      config = '{}-{}-{}'.format(
          filter(str.isalnum, std_flag), exceptions, build_type)
      build_dir = 'build-{}'.format(config)
      os.mkdir(build_dir)
      os.chdir(build_dir)
      result[config] = { 'Configure': None, 'Build': None, 'Test': None }

      tests = os.environ['TESTS'].split()
      if std_flag.endswith(('11', '1y', '14', '1z')) and 'libc++' in tests:
        tests.remove('libc++')

      result[config]['Configure'] = subprocess.call([
        'cmake', '-GNinja',
                 '-DCMAKE_BUILD_TYPE={}'.format(build_type),
                 '-DMPARK_VARIANT_EXCEPTIONS={}'.format(exceptions),
                 '-DMPARK_VARIANT_INCLUDE_TESTS={}'.format(';'.join(tests)),
                 '..',
      ])
      if result[config]['Configure'] == 0:
        result[config]['Build'] = subprocess.call([
          'cmake', '--build', '.', '--', '-k', '0'])
        if result[config]['Build'] == 0:
          result[config]['Test'] = subprocess.call([
            'ctest', '--output-on-failure'])
      os.chdir('..')

pprint.pprint(result)
exit(any(status != 0 for d in result.itervalues() for status in d.itervalues()))



# distutils: language=c++

from libcpp cimport bool
from libcpp.string cimport string


cdef extern from "boutexception_helper.hxx":
     cdef void raise_bout_py_error()


cdef extern from "bout/options_io.hxx" namespace "bout":
    cdef void writeDefaultOutputFile(Options& options);
    cppclass OptionsIO:
        @staticmethod
        OptionsIO * create(string filename)
        Options read();
        void write(const Options& options);
        void write(const Options& options, string time_dim);
        void verifyTimesteps() const;


cdef extern from "bout/options.hxx":
    cppclass Options:
        Options()
        @staticmethod
        Options* getRoot()
        @staticmethod
        Options& root()
        Options& operator[](string)
        T operator=[T](T)
        T as[T]() const
        void assign[T](T val)
        void assign[T](T val, string souce)
        void force[T](T val)
        void force[T](T val, string souce)
        void assignRepeat[T](T val)
        void assignRepeat[T](T val, string time_dimension, bool save_repeat, string source)
        bool isSet()
        T withDefault[T](T default)
        Options* getSection(string section)
        void set(string, string, string, bool)
        void get(string, string&, string)
        void get(string, double&, double)
        void get(string, bool&, bool)
        void cleanCache()
        void setConditionallyUsed()


cdef extern from "bout/optionsreader.hxx":
    cppclass OptionsReader:
        @staticmethod
        OptionsReader *getInstance()
        void read(Options *options, const char *file, ...);



#!/usr/bin/env python3

import argparse
import copy
import difflib
import re

format_replacements = {
    "c": "c",
    "d": "d",
    "e": "e",
    "f": "f",
    "g": "g",
    "i": "d",
    "ld": "d",
    "le": "e",
    "lu": "d",
    "p": "p",
    "s": "s",
    "zu": "d",
}


def fix_format_replacement(format_replacement, source):
    """Replace printf format with fmt format"""
    return re.sub(
        r"%([0-9]*\.?[0-9]*){}".format(format_replacement[0]),
        r"{{:\1{}}}".format(format_replacement[1]),
        source,
    )


def fix_trivial_format(source):
    """Reduce trivial formatting of strings to just the string"""

    def trivial_replace(match):
        if match.group(2):
            return "{}{}{}".format(match.group(1), match.group(2), match.group(4))
        if match.group(3):
            return "{}{}{}".format(match.group(1), match.group(3), match.group(4))
        raise ValueError("Found an unexpected match: {}".format(match))

    return re.sub(
        r"""
        (.*)?
        "{:s}",\s*                  # Entire format is just a string
        (?:([\w_]+)\.c_str\(\)      # And replacement is std::string::c_str
        |(".*?"))
        (.*)?
        """,
        trivial_replace,
        source,
        flags=re.VERBOSE,
    )


def fix_string_c_str(source):
    """Fix formats that use {:s} where the replacement is using std::string::c_str"""
    return re.sub(
        r"""
        (".*{:s}[^;]*?",)       # A format string containing {:s}
        \s*([^;]+?)\.c_str\(\) # Replacement of std::string::c_str
        """,
        r"\1 \2",
        source,
        flags=re.VERBOSE,
    )


def fix_trace(source):
    """Fix TRACE macros where fix_string_c_str has failed for some reason"""
    return re.sub(
        r"""
        (TRACE\(".*{:s}.*",)
        \s*([\w_]+)\.c_str\(\)\); # Replacement of std::string::c_str
        """,
        r"\1 \2);",
        source,
        flags=re.VERBOSE,
    )


def fix_toString_c_str(source):
    """Fix formats that call toString where the replacement is using std::string::c_str"""
    return re.sub(
        r"""
        (".*{:s}[^;]*?",.*?)         # A format string containing {:s}
        (toString\(.*?\))\.c_str\(\) # Replacement of std::string::c_str
        """,
        r"\1\2",
        source,
        flags=re.VERBOSE,
    )


def apply_fixes(format_replacements, source):
    """Apply the various fixes for each factory to source. Returns
    modified source

    Parameters
    ----------
    factories
        Dictionary of factory properties
    source
        Text to update
    """

    modified = source

    for format_replacement in format_replacements.items():
        modified = fix_format_replacement(format_replacement, modified)

    modified = fix_trivial_format(modified)
    modified = fix_string_c_str(modified)
    modified = fix_trace(modified)

    return modified


def yes_or_no(question):
    """Convert user input from yes/no variations to True/False"""
    while True:
        reply = input(question + " [y/N] ").lower().strip()
        if not reply or reply[0] == "n":
            return False
        if reply[0] == "y":
            return True


def create_patch(filename, original, modified):
    """Create a unified diff between original and modified"""

    patch = "\n".join(
        difflib.unified_diff(
            original.splitlines(),
            modified.splitlines(),
            fromfile=filename,
            tofile=filename,
            lineterm="",
        )
    )

    return patch


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Fix format specifiers")

    parser.add_argument("files", action="store", nargs="+", help="Input files")
    parser.add_argument(
        "--force", "-f", action="store_true", help="Make changes without asking"
    )
    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Don't print patches"
    )
    parser.add_argument(
        "--patch-only", "-p", action="store_true", help="Print the patches and exit"
    )

    args = parser.parse_args()

    if args.force and args.patch_only:
        raise ValueError("Incompatible options: --force and --patch")

    for filename in args.files:
        with open(filename, "r") as f:
            contents = f.read()
        original = copy.deepcopy(contents)

        modified = apply_fixes(format_replacements, contents)
        patch = create_patch(filename, original, modified)

        if args.patch_only:
            print(patch)
            continue

        if not patch:
            if not args.quiet:
                print("No changes to make to {}".format(filename))
            continue

        if not args.quiet:
            print("\n******************************************")
            print("Changes to {}\n".format(filename))
            print(patch)
            print("\n******************************************")

        if args.force:
            make_change = True
        else:
            make_change = yes_or_no("Make changes to {}?".format(filename))

        if make_change:
            with open(filename, "w") as f:
                f.write(modified)



from __future__ import print_function
from builtins import str

#
# Generate the test case using SymPy
#
# Equations are:
#
# df/dt = DDX(g)
# dg/dt = DDX(f)
#
#

from sympy import symbols, cos, sin, diff

# Define symbols

x = symbols("x")
t = symbols("t")

# Define the manufactured solution

f = 0.9 + 0.9 * x + 0.2 * cos(10 * t) * sin(5.0 * x**2)
g = 0.9 + 0.7 * x + 0.2 * cos(7 * t) * sin(2.0 * x**2)

# Calculate gradients in x for boundaries

dfdx = diff(f, x)
dgdx = diff(g, x)

# Calculate sources

Sf = diff(f, t) - dgdx
Sg = diff(g, t) - dfdx

#######################

print("F:")
print("solution = " + str(f))
print("d/dx     = " + str(dfdx))
print("source   = " + str(Sf))

print("\nG:")
print("solution = " + str(g))
print("d/dx     = " + str(dgdx))
print("source   = " + str(Sg))



from __future__ import print_function

# Test
from sympy import symbols, cos, diff
from numpy import pi as pin
from boutdata.mms import exprToStr

# Define symbols for orthogonal (theta, phi) coordinates

theta = symbols("theta")
phi = symbols("phi")
psi = symbols("psi")
t = symbols("t")
pi = symbols("pi")

# Define solution (theta, phi)

# f = cos(theta + phi - t)**2
f = cos(psi + theta + phi - t) ** 2

# Calculate solution in (theta, phi)

dfdt = diff(f, theta) + diff(f, phi) + diff(f, psi)
dfdpsi = diff(f, psi)
dfdtheta = diff(f, theta)

# Source

Sf = diff(f, t) - dfdt

# Define new coordinate system
#
# y = theta
# z = phi - zShift
#
# where
#
# zShift = int_{theta_0}^{theta} nu(theta) dtheta
#
# and nu(theta) is the field-line pitch
#

x = symbols("x")
y = symbols("y")
z = symbols("z")

zShift = 0.1 * theta  # *psi
nu = diff(zShift, theta)
I = diff(zShift, psi)
H = diff(zShift, theta)
shiftangle = zShift.subs([(theta, 2 * pin)]) - zShift.subs([(theta, 0.0)])

yShift = 0.2 * psi  # 0.05*(0.5-psi)*sin(theta) #0.2*psi #
eta = diff(yShift, psi)
G = 1 - diff(yShift, theta)

replace = [(phi, z + zShift), (theta, y + yShift), (psi, x)]  # z = phi - zShift
# replace = [ (phi, z + zShift), (psi, x) ]    # z = phi - zShift

# Replace and print sources, solutions etc.

print("[mesh]")
print("zShift = " + exprToStr(zShift.subs(replace)))
print("nu = " + exprToStr(nu.subs(replace)))
print("shiftangle = " + exprToStr(shiftangle.subs(replace)))
print("I = " + exprToStr(I.subs(replace)))
print("HH = " + exprToStr(H.subs(replace)))
print("")
print("eta = " + exprToStr(eta.subs(replace)))
print("yShift = " + exprToStr(yShift.subs(replace)))
print("G = " + exprToStr(G.subs(replace)))


print("\n[f]")
print("solution = " + exprToStr(f.subs(replace)))
print("\nsource = " + exprToStr(Sf.subs(replace)))
print("dfdpsi = " + exprToStr(dfdpsi.subs(replace)))
print("dfdtheta = " + exprToStr(dfdtheta.subs(replace)))



#
# SNB model example
#
# Uses a step in the temperature, intended for comparison to VFP results

length = 6e-4  # Domain length in m

qe = 1.602176634e-19

import numpy as np
import matplotlib.pyplot as plt

from boututils.run_wrapper import build_and_log, launch_safe
from boutdata.collect import collect

path = "step"

build_and_log("Step SNB")
# Run the case
s, out = launch_safe("./conduction-snb -d " + path, nproc=1, mthread=1, pipe=True)

Te = collect("Te", path=path).ravel()

ny = len(Te)
dy = length / ny

position = (np.arange(ny) + 0.5) * length / ny

# Read divergence of heat flux
div_q = collect("Div_Q", path=path).ravel()
div_q_SH = collect("Div_Q_SH", path=path).ravel()

# Integrate the divergence of flux to get heat flux W/m^2
q = np.cumsum(div_q) * qe * dy
q_SH = np.cumsum(div_q_SH) * qe * dy

# Subtract the minimum value of each heat flux
q -= np.amin(q)
q_SH -= np.amin(q_SH)

# Read reference values
Te_ref = np.loadtxt("temperature.csv", delimiter=",")
Te_ref[:, 0] *= 1e-4  # Convert X axis to m

SH_ref = np.loadtxt("spitzer-harm.csv", delimiter=",")
SH_ref[:, 0] *= 1e-4

SNB_ref = np.loadtxt("snb.csv", delimiter=",")
SNB_ref[:, 0] *= 1e-4

VFP_ref = np.loadtxt("vfp.csv", delimiter=",")
VFP_ref[:, 0] *= 1e-4

#########################################

fig, ax1 = plt.subplots()

color = "tab:red"
ax1.plot(position, Te * 1e-3, color=color, label="Te")
ax1.plot(Te_ref[:, 0], Te_ref[:, 1], color=color, marker="o", label="Reference Te")
ax1.set_xlabel("position [m]")
ax1.set_ylabel("Electron temperature [keV]", color=color)
ax1.set_ylim(0, 1)
ax1.tick_params(axis="y", colors=color)

ax2 = ax1.twinx()
ax2.plot(position, q_SH * 1e-4, "-k", label="Spitzer-Harm")
ax2.plot(SH_ref[:, 0], SH_ref[:, 1], "--k", label="Reference SH")

ax2.plot(position, q * 1e-4, "-b", label="SNB")
ax2.plot(SNB_ref[:, 0], SNB_ref[:, 1], "--b", label="Reference SNB")

ax2.plot(VFP_ref[:, 0], VFP_ref[:, 1], "--g", label="Reference VFP")

ax2.set_ylabel("Heat flux W/cm^2")
ax2.set_ylim(bottom=0.0)

plt.legend()
fig.tight_layout()

plt.xlim(0, 3.5e-4)

plt.savefig("snb-step.png")
plt.show()



#!/usr/bin/env python3

from __future__ import print_function
from __future__ import division

from builtins import str, range

from numpy import *
from scipy.integrate import quad

from boututils.datafile import DataFile

######################################################

nx = 68  # Number of radial grid points
ny = 32  # Number of poloidal (parallel) grid points

varyBp = False

output = "cyclone_" + str(nx) + "x" + str(ny) + ".nc"

######################################################

Ni = 1.0  # Ion density in 10^20 m^-3
Ti = 1000  # Temperature in eV (Te = Ti)
Rmaj = 4  # Major radius [meters]
q = 1.4  # Safety factor q = r*Bt/(R*Bp)
s = 0.776  # Magnetic shear s = (r/q) dq/dr
eta_i = 3.114  # Ratio of density to temp. length scales eta = L_n / L_T
epsilon = 0.18  # Inverse aspect ratio epsilon = r / R
Rnorm = 6.92  # Ratio of major radius to L_T  Rnorm  = R / L_T
rho_norm = 0.01  # Normalised ion gyro-radius rho_norm = rho_i / L_T
r_wid = 100  # Radial extent, normalised to gyro-radius r_wid = dr / rho_i

Mi = 2.0 * 1.67262158e-27  # Ion mass [kg]. Deuterium

######################################################


def eps_integral(eps, theta=None):
    if theta == None:
        theta = 2.0 * pi
    return (quad(lambda t: 1.0 / ((1.0 - eps * cos(t)) ** 2), 0.0, theta))[0]


rminor = Rmaj * epsilon  # Minor radius [m]
L_T = Rmaj / Rnorm  # Temp. length scale [m]
L_n = eta_i * L_T  # Density length scale [m]
rho_i = rho_norm * L_T  # Ion Larmor radius [m]
Bt0 = sqrt(2.0 * Ti * Mi / 1.602e-19) / rho_i  # Toroidal field from rho_i [T]
Bp = rminor * Bt0 * eps_integral(epsilon) / (2.0 * pi * q * Rmaj)  # Poloidal field [T]

dr = r_wid * rho_i  # Width of domain [m]

theta = 2.0 * pi * arange(0, float(ny)) / float(ny)

Rxy = zeros([nx, ny])
Zxy = Rxy.copy()
for i in range(ny):
    Rxy[:, i] = Rmaj - rminor * cos(theta[i])
    Zxy[:, i] = rminor * sin(theta[i])

dy = zeros([nx, ny]) + 2.0 * pi / float(ny)
hthe = zeros([nx, ny]) + rminor

Btxy = Bt0 * Rmaj / Rxy

print(
    "Toroidal field varies from "
    + str(Bt0 * Rmaj / (Rmaj + rminor))
    + " to "
    + str(Bt0 * Rmaj / (Rmaj - rminor))
)

# Minor radius offset
drprof = dr * ((arange(nx) / float(nx - 1)) - 0.5)

# q profile
qprof = q + (s * q / rminor) * drprof

print("q varies from " + str(min(qprof)) + " to " + str(max(qprof)))

ShiftAngle = qprof * 2.0 * pi

Bpxy = zeros([nx, ny])
if varyBp:
    # Vary Bp to get shear
    for y in range(ny):
        Bpxy[:, y] = Bp * q / qprof
    print("Poloidal field varies from " + str(amin(Bpxy)) + " to " + str(amax(Bpxy)))
else:
    # Constant Bp, but shift angle varies
    Bpxy += Bp

dx = Bp * (dr / float(nx - 1)) * Rxy

Bxy = sqrt(Btxy**2 + Bpxy**2)

zShift = zeros([nx, ny])
qint = eps_integral(epsilon)

for i in range(1, ny):
    zShift[:, i] = ShiftAngle * eps_integral(epsilon, theta=theta[i]) / qint

# Make zShift = 0 on outboard midplane (for plotting mainly)
y0 = int(ny / 2)
zs0 = zShift[:, y0]
for i in range(ny):
    zShift[:, i] -= zs0

Ni0 = zeros([nx, ny])
Ti0 = Ni0
for i in range(ny):
    Ni0[:, i] = Ni * exp(-drprof / L_n)
    Ti0[:, i] = Ti * exp(-drprof / L_T)
Te0 = Ti0

pressure = Ni0 * (Ti0 + Te0) * 1.602e-19 * 1.0e20  # In Pascals

Jpar0 = zeros([nx, ny])

# Shape       : Rxy, Zxy
# Differencing: hthe, dx, dy
# Profiles    : Ni0, Ti0, Te0, pressure, Jpar0
# B field     : Btxy, Bpxy, Bxy
# q profile   : qprof

######################################################
# Curvature

# Bxy is constant in x, so need to supply logB too

logB = zeros([nx, ny])

for x in range(nx):
    for y in range(ny):
        rpos = (float(x) / float(nx - 1) - 0.5) * dr
        R = Rmaj - (rminor + rpos) * cos(theta[y])
        Bt = Bt0 * Rmaj / R
        logB[x, y] = log(sqrt(Bt**2 + Bp**2))

######################################################
# Topology: Just in the core

ixseps1 = nx
ixseps2 = nx
jyseps1_1 = -1
jyseps1_2 = int(ny / 2)
jyseps2_1 = jyseps1_2
jyseps2_2 = ny - 1
ny_inner = jyseps1_2

# Only one region
yup_xsplit = [nx]
ydown_xsplit = [nx]
yup_xin = [0]
yup_xout = [-1]
ydown_xin = [0]
ydown_xout = [-1]
nrad = [nx]
npol = [ny]

######################################################

print("Writing grid to file " + output)

of = DataFile()
of.open(output, create=True)

of.write("nx", nx)
of.write("ny", ny)

# Topology for original scheme
of.write("ixseps1", ixseps1)
of.write("ixseps2", ixseps2)
of.write("jyseps1_1", jyseps1_1)
of.write("jyseps1_2", jyseps1_2)
of.write("jyseps2_1", jyseps2_1)
of.write("jyseps2_2", jyseps2_2)
of.write("ny_inner", ny_inner)

# Grid spacing
of.write("dx", dx)
of.write("dy", dy)

of.write("ShiftAngle", ShiftAngle)
of.write("zShift", zShift)

of.write("Rxy", Rxy)
of.write("Zxy", Zxy)
of.write("Bpxy", Bpxy)
of.write("Btxy", Btxy)
of.write("Bxy", Bxy)
of.write("hthe", hthe)

# Topology for general configurations
of.write("yup_xsplit", yup_xsplit)
of.write("ydown_xsplit", ydown_xsplit)
of.write("yup_xin", yup_xin)
of.write("ydown_xin", ydown_xin)
of.write("ydown_xout", ydown_xout)
of.write("nrad", nrad)
of.write("npol", npol)

# plasma profiles
of.write("pressure", pressure)
of.write("Jpar0", Jpar0)
of.write("Ni0", Ni0)
of.write("Te0", Te0)
of.write("Ti0", Ti0)
of.write("Ni_x", Ni)
of.write("Te_x", Ti)
of.write("Ti_x", Ti)
of.write("bmag", Bt0)
of.write("rmag", Rmaj)

# Curvature
of.write("logB", logB)

of.close()

print("Done")



#!/usr/bin/env python3

from sympy import Symbol, Eq
from sympy.matrices import Matrix
from sympy.printing import ccode
from sympy.simplify import combsimp as simp
from sympy.utilities.codegen import codegen


def pow(a, b):
    if b == 0:
        return "1"
    if b == 1:
        return a
    else:
        return "%s**%d" % (a, b)


def factorial(a):
    if a == 0 or a == 1:
        return 1
    else:
        assert a > 0
        return a * factorial(a - 1)


def gen_code(order, matrix_type):
    x = [Symbol("spacing.f%d" % i) for i in range(order)]
    matrix = matrix_type(x)
    A = Matrix(order, order, matrix)

    try:
        iA = A.inv()
    except:
        import sys

        print(A, matrix, file=sys.stderr)
        raise
    ret = ""
    for i in range(order):
        ret += ccode(simp(iA[0, i]), assign_to="facs.f%d" % i)
        ret += "\n"
    return ret


def taylor(x, i, j):
    if j >= 0:
        return x[i] ** j / factorial(j)
    else:
        return 0


class dirichlet:
    name = "dirichlet"

    def __init__(self, x):
        self.x = x

    def __call__(self, i, j):
        return taylor(self.x, i, j)


class neumann:
    name = "neumann"

    def __init__(self, x):
        self.x = x

    def __call__(self, i, j):
        if i == 0:
            return taylor(self.x, i, j - 1)
        else:
            return taylor(self.x, i, j)


if __name__ == "__main__":
    print(gen_code(3, dirichlet))



from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div

from numpy import *

# from scipy.io import readsav;
import matplotlib.pyplot as plt

# Dynamic matplotlib settings
from matplotlib import rcParams

rcParams["font.size"] = 20
rcParams["legend.fontsize"] = "small"
rcParams["legend.labelspacing"] = 0.1
rcParams["lines.linewidth"] = 2
rcParams["savefig.bbox"] = "tight"
# Create image directory if not exists
import os

if not os.path.exists("image"):
    os.makedirs("image")

# fphi = transpose(readsav('fphi.idl.dat')['fphi'])[:,:,:,];
fphi = load("fp.npy")

plt.figure()
for i in range(1, 9):
    print("Growth rate for mode number", i)
    print(gradient(log(abs(fphi[34, 32, i, :]))))
    plt.semilogy((abs(fphi[34, 32, i, :])), label="n=" + str(i * 5))

plt.legend(loc=2)
plt.xlabel("Time")
plt.savefig("image/plotmode.png")
plt.savefig("image/plotmode.eps")
plt.show(block=False)
plt.figure()
for i in range(1, 9):
    plt.plot(abs(fphi[:, 32, i, -1]), label="n=" + str(i * 5))

plt.legend()
plt.xlabel("X index")
plt.savefig("image/plotmodeamp.png")
plt.savefig("image/plotmodeamp.eps")
plt.show(block=False)
plt.figure()
for i in range(1, 9):
    plt.plot(
        old_div(abs(fphi[:, 32, i, -1]), abs(fphi[:, 32, i, -1]).max()),
        label="n=" + str(i * 5),
    )

plt.legend()
plt.xlabel("X index")
plt.savefig("image/plotmodenorm.png")
plt.savefig("image/plotmodenorm.eps")
plt.show()



import numpy as np
from boutdata.collect import collect
import pickle

try:
    from past.utils import old_div
except ImportError:

    def old_div(a, b):
        return a / b


def blob_velocity(n, **kwargs):
    from boututils import calculus as Calc
    # Calculate blob velocity in normalized time and normalized grid spacing
    #
    # Input: Blob density as a 3D vector in the form  n[t,x,z] where t is time and x,z are the perpendicular spatial coordinates
    #
    # Keywords:
    #
    #           type='peak' -> Calculate velocity of the peak density
    #           type='COM' -> Calculate centre of mass velocity
    #           Index=True -> return indices used to create velocity
    #
    # Default: Peak velocity with no index returning

    size = n.shape

    try:
        v_type = kwargs["type"]
    except:
        v_type = "peak"  # Default to peak velocity calculation
    try:
        return_index = kwargs["Index"]
    except:
        return_index = False  # Default to no index returning

    if v_type == "peak":
        x = np.zeros(size[0])
        z = np.zeros(size[0])
        for i in np.arange(size[0]):
            nmax, nmin = np.amax((n[i, :, :])), np.amin((n[i, :, :]))
            xpos, zpos = np.where(n[i, :, :] == nmax)
            x[i] = xpos[0]
            z[i] = zpos[0]

    if v_type == "COM":
        x = np.zeros(size[0])
        z = np.zeros(size[0])
        for i in np.arange(size[0]):
            data = n[i, :, :] - n[0, 0, 0]  # use corner cell rather than nmin
            ntot = np.sum(data[:, :])

            z[i] = old_div(
                np.sum(np.sum(data[:, :], axis=0) * (np.arange(size[2]))), ntot
            )
            x[i] = old_div(
                np.sum(np.sum(data[:, :], axis=1) * (np.arange(size[1]))), ntot
            )

    vx = Calc.deriv(x)
    vz = Calc.deriv(z)

    if return_index:
        return vx, vz, x, z
    else:
        return vx, vz


data = "data"

if True:
    import sys

    if len(sys.argv) > 1:
        data = sys.argv[1]

n = collect("n", path=data, info=False)


vx, vy, xx, yy = blob_velocity(n[:, :, 0, :], type="COM", Index=True)

f = open("Velocity.dat", "wb")
pickle.dump(vx, f)
f.close()

f = open("Position.dat", "wb")
pickle.dump(xx, f)
f.close()


f = open("Velocity.dat", "rb")
vx = pickle.load(f)
f.close()

try:
    import matplotlib.pyplot as plt

    plt.plot(vx)
    plt.show()
except ImportError:
    pass



#!/usr/bin/env python3

#
# Generate input mesh for slab shear Alfven wave case
#
############################################################
# Slab inputs

from numpy import linspace, ndarray, sqrt, arange, amin, amax, pi
from boututils.datafile import DataFile

# Size of the domain
Lx = 0.1  # Radial domain size [m]
Ltor = 10.0  # "Toroidal" length [m]
Lpol = 1.0  # "Poloidal" length [m]

Bt = 1.0  # Magnetic field [T]
Bp = 0.1  # Poloidal field at the middle of the domain [T]
Bpprime = 1.0  # Bp gradient [T/m] Bp(x) = Bp + Bpprime * x

# Number of grid points in each direction
nx = 34
ny = 64
nz = 64  # Note: Never used here

ymid = 32  # Index of the midplane where X-Z mesh is orthogonal

############################################################

# Effective major radius
R = Ltor / (2.0 * pi)

# Effective minor radius
hthe = Lpol / (2.0 * pi)

# Safety factor
q = Bt * hthe / (Bp * R)
print("Safety factor: %e" % q)

# Magnetic shear
s = (-q / Bp) * Bpprime * (hthe / q)
print("Magnetic shear: %e" % s)

# Normalised x coordinate from 0 to 1
x = linspace(0, 1, nx)

# Set poloidal magnetic field
Bpx = Bp + (x - 0.5) * Lx * Bpprime

Bpxy = ndarray([nx, ny])
for x in range(nx):
    Bpxy[x, :] = Bpx[x]

Bxy = sqrt(Bpxy**2 + Bt**2)

# Calculate change in poloidal flux
dr = Lx / nx  # Constant mesh spacing in radius
dx = Bpxy * R * dr  # Slightly non-uniform mesh in x, since Bp*R is not constant

# Poloidal angle
dy = 2.0 * pi / ny

# Calculate zShift and magnetic shear.
# Since metrics vary only in x, integration is trivial
zShift = ndarray([nx, ny])
sinty = ndarray([nx, ny])
TwistShift = ndarray(nx)

for x in range(nx):
    qsafe = Bt * hthe / (Bpxy[x, 0] * R)  # safety factor
    zShift[x, :] = qsafe * (arange(ny) - ymid) * dy
    TwistShift[x] = qsafe * 2.0 * pi  # Angle to use for twist-shift condition

    # dq/dx = dq/dr * dr/dx
    dqdx = (-qsafe * Bpprime / Bpxy[x, 0]) * (dr / dx[x, 0])

    # Integrate to get shear
    sinty[x, :] = dqdx * (arange(ny) - ymid) * dy

print(
    "Safety factor varies between %e and %e"
    % (amin(TwistShift) / (2.0 * pi), amax(TwistShift) / (2.0 * pi))
)
############################################################
# Write grid file

with DataFile("slab.grd.nc", create=True) as d:
    d.write("nx", nx)
    d.write("ny", ny)

    d.write("dx", dx)
    d.write("dy", dy)

    d.write("Rxy", R)
    d.write("hthe", hthe)
    d.write("Bpxy", Bpxy)
    d.write("Btxy", Bt)
    d.write("Bxy", Bxy)

    d.write("sinty", sinty)
    d.write("zShift", zShift)
    d.write("ShiftAngle", TwistShift)



from __future__ import print_function
from builtins import str

#
# Generate the test case using SymPy
#
# Equations are:
#
# d/dt(n) = Dx * D2DX2(n) + Dy * D2DY2(n) + Dz * D2DZ2(n)
#

from sympy import symbols, cos, sin, diff

####


def bracket(f, g):
    """
    Calculates [f,g] symbolically
    """

    dfdx = diff(f, x)
    dfdz = diff(f, z)

    dgdx = diff(g, x)
    dgdz = diff(g, z)

    return dfdz * dgdx - dfdx * dgdz


def DDZ(f):
    return diff(f, z)


def Delp2(f):
    """Laplacian in X-Z"""
    d2fdx2 = diff(f, x, 2)
    d2fdz2 = diff(f, z, 2)

    return d2fdx2 + d2fdz2


def Delp4(f):
    d4fdx4 = diff(f, x, 4)
    d4fdz4 = diff(f, z, 4)

    return d4fdx4 + d4fdz4


def D2DX2(f):
    return diff(f, x, 2)


def D2DY2(f):
    return diff(f, y, 2)


def D2DZ2(f):
    return diff(f, z, 2)


def exprToStr(expr):
    """Convert a sympy expression to a string for BOUT++ input"""
    tmp = str(expr).replace("**", "^")  # Replace exponent operator
    tmp = tmp.replace("xl", "N:xl")
    return tmp.replace("yl", "N:yl")


####

# Parameters
Dx = 1.0
Dy = 1.0
Dz = 1.0

# Define symbols

x = symbols("x")
y = symbols("y")
xl = symbols("xl")
yl = symbols("yl")
z = symbols("z")
t = symbols("t")
pi = symbols("pi")

# Define the manufactured solution

n = (
    0.9
    + 0.9 * x
    + 0.2 * sin(5.0 * x**2 - 2 * z)
    + cos(10 * z) * cos(y) * sin(y * 7 + 1.234)
)


# Calculate gradients for boundaries

dndx = diff(n, x)
dndy = diff(n, y)

# Calculate RHS function

dndt = Dx * D2DX2(n) + Dy * D2DY2(n) + Dz * D2DZ2(n)

# Calculate sources

Sn = diff(n, t) - dndt

# x and y-domains are Lx and Ly long, respectively
# change to corresponding coordinates. Standard BOUT++ coordinates
# have Lx = 1, Ly = 2pi.
# Scaling takes place in BOUT.inp
scale_coordinates = [(x, xl), (y, yl)]
n = n.subs(scale_coordinates)
dndx = dndx.subs(scale_coordinates)
dndy = dndy.subs(scale_coordinates)
Sn = Sn.subs(scale_coordinates)
#######################

print("[n]")
print("solution = " + exprToStr(n))
print("\nddx     = " + exprToStr(dndx))
print("\nddy     = " + exprToStr(dndy))
print("\nsource   = " + exprToStr(Sn))



import csv


def read_file(filename):
    reader = csv.reader(open(filename, "r"), delimiter="\t", skipinitialspace=True)

    # Skip header
    for _, _ in zip(range(4), reader):
        continue

    case_lines = {}
    for line in reader:
        if line == []:
            break
        case_lines[line[0].rstrip(".")] = line[1]

    titles = next(reader)
    cases_weak = {col.strip(): [] for col in titles[:-1]}

    for line in reader:
        if line == []:
            break
        for title, col in zip(titles, line[:-1]):
            cases_weak[title].append(float(col))

    return case_lines, cases_weak, titles



#!/usr/bin/env python3

import argparse
import copy
import difflib
import re
import textwrap

# List of macros, their replacements and what header to find them
# in. Each element should be a dict with "old", "new" and "headers"
# keys, with "old" and "new" values being strings, and "headers" being a
# list of strings. "new" can also be None if the macro has been removed, which
# will cause an error to be printed if the macro is found.
MACRO_REPLACEMENTS = [
    {
        "old": "REVISION",
        "new": "bout::version::revision",
        "headers": ["bout/revision.hxx"],
        "macro": False,
        "always_defined": True,
    },
    {
        "old": "BOUT_VERSION_DOUBLE",
        "new": "bout::version::as_double",
        "headers": ["bout/version.hxx", "bout.hxx"],
        "macro": False,
        "always_defined": True,
    },
    {
        "old": "BOUT_VERSION_STRING",
        "new": "bout::version::full",
        "headers": ["bout/version.hxx", "bout.hxx"],
        "macro": False,
        "always_defined": True,
    },
    # Next one is not technically a macro, but near enough
    {
        "old": "BOUT_VERSION",
        "new": "bout::version::full",
        "headers": ["bout/version.hxx", "bout.hxx"],
        "macro": False,
        "always_defined": True,
    },
    {
        "old": "BACKTRACE",
        "new": "BOUT_USE_BACKTRACE",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_ARKODE",
        "new": "BOUT_HAS_ARKODE",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_CVODE",
        "new": "BOUT_HAS_CVODE",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_HDF5",
        "new": None,
        "headers": [],
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_IDA",
        "new": "BOUT_HAS_IDA",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_LAPACK",
        "new": "BOUT_HAS_LAPACK",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "LAPACK",
        "new": "BOUT_HAS_LAPACK",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_NETCDF",
        "new": "BOUT_HAS_NETCDF",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_PETSC",
        "new": "BOUT_HAS_PETSC",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_PRETTY_FUNCTION",
        "new": "BOUT_HAS_PRETTY_FUNCTION",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HAS_PVODE",
        "new": "BOUT_HAS_PVODE",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "TRACK",
        "new": "BOUT_USE_TRACK",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "NCDF4",
        "new": "BOUT_HAS_NETCDF",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "NCDF",
        "new": "BOUT_HAS_LEGACY_NETCDF",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "HDF5",
        "new": None,
        "headers": [],
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "DEBUG_ENABLED",
        "new": "BOUT_USE_OUTPUT_DEBUG",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "BOUT_FPE",
        "new": "BOUT_USE_SIGFPE",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "LOGCOLOR",
        "new": "BOUT_USE_COLOR",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
    {
        "old": "OPENMP_SCHEDULE",
        "new": "BOUT_OPENMP_SCHEDULE",
        "headers": "bout/build_config.hxx",
        "macro": True,
        "always_defined": True,
    },
]


def fix_include_version_header(old, headers, source):
    """Make sure version.hxx header is included"""

    if not isinstance(headers, list):
        headers = [headers]

    # If header is already included, we can skip this fix
    for header in headers:
        if (
            re.search(r'^#\s*include.*(<|"){}(>|")'.format(header), source, flags=re.M)
            is not None
        ):
            return source

    # If the old macro isn't in the file, we can skip this fix
    if re.search(r"\b{}\b".format(old), source) is None:
        return source

    # Now we want to find a suitable place to stick the new include
    # Good candidates are includes of BOUT++ headers
    includes = []
    source_lines = source.splitlines()
    for linenumber, line in enumerate(source_lines):
        if re.match(r"^#\s*include.*bout/", line):
            includes.append(linenumber)
        if re.match(r"^#\s*include.*physicsmodel", line):
            includes.append(linenumber)

    if includes:
        last_include = includes[-1] + 1
    else:
        # No suitable includes, so just stick at the top of the file
        last_include = 0
    source_lines.insert(last_include, '#include "{}"'.format(headers[0]))

    return "\n".join(source_lines)


def fix_ifdefs(old, source):
    """Remove any code inside #ifdef/#ifndef blocks that would now not be compiled"""
    source_lines = source.splitlines()

    # Something to keep track of nested sections
    in_ifdef = None
    # List of (#ifdef or #ifndef, dict of start/else/end lines)
    macro_blocks = []
    for linenumber, line in enumerate(source_lines):
        if_def = re.match(r"#\s*(ifn?def)\s*(.*)", line)
        else_block = re.match(r"#\s*else", line)
        endif = re.match(r"#\s*endif", line)
        if not (if_def or else_block or endif):
            continue
        # Now we need to keep track of whether we're inside an
        # interesting #ifdef/ifndef, as they might be nested, and we
        # want to find the matching #endif and #else
        if endif:
            if in_ifdef is not None:
                in_ifdef -= 1
            if in_ifdef == 0:
                in_ifdef = None
                macro_blocks[-1]["end"] = linenumber
            continue
        if else_block:
            if in_ifdef == 1:
                macro_blocks[-1]["else"] = linenumber
            continue
        if if_def.group(2) == old:
            in_ifdef = 1
            macro_blocks.append({"start": linenumber, "if_def_type": if_def.group(1)})
        elif in_ifdef is not None:
            in_ifdef += 1

    if macro_blocks == []:
        return source

    # Get all of the lines to be removed
    lines_to_remove = set()
    for block in macro_blocks:
        lines_to_remove |= set(block.values())
        if block["if_def_type"] == "ifdef":
            if "else" in block:
                # Delete the #else block for #ifdef
                lines_to_remove |= set(range(block["else"], block["end"]))
        else:
            # Keep the #else block for #ifndef if there is one, otherwise remove the
            # whole block
            lines_to_remove |= set(
                range(block["start"], block.get("else", block["end"]))
            )

    # Apparently this is actually the best way of removing a bunch of (possibly)
    # non-contiguous indices
    modified_lines = [
        line for num, line in enumerate(source_lines) if num not in lines_to_remove
    ]

    return "\n".join(modified_lines)


def fix_always_defined_macros(old, new, source):
    """Fix '#ifdef's that should become plain '#if'"""
    new_source = re.sub(r"#ifdef\s+{}\b".format(old), r"#if {}".format(new), source)
    return re.sub(r"#ifndef\s+{}\b".format(old), r"#if !{}".format(new), new_source)


def fix_replacement(old, new, source):
    """Straight replacements"""
    return re.sub(r'([^"_])\b{}\b([^"_])'.format(old), r"\1{}\2".format(new), source)


def apply_fixes(replacements, source):
    """Apply all fixes in this module"""
    modified = copy.deepcopy(source)

    for replacement in replacements:
        if replacement["new"] is None:
            print(
                "'%s' has been removed, please delete from your code"
                % replacement["old"]
            )
            continue

        modified = fix_include_version_header(
            replacement["old"], replacement["headers"], modified
        )
        if replacement["macro"] and replacement["always_defined"]:
            modified = fix_always_defined_macros(
                replacement["old"], replacement["new"], modified
            )
        elif replacement["always_defined"]:
            modified = fix_ifdefs(replacement["old"], modified)
        modified = fix_replacement(replacement["old"], replacement["new"], modified)

    return modified


def yes_or_no(question):
    """Convert user input from yes/no variations to True/False"""
    while True:
        reply = input(question + " [y/N] ").lower().strip()
        if not reply or reply[0] == "n":
            return False
        if reply[0] == "y":
            return True


def create_patch(filename, original, modified):
    """Create a unified diff between original and modified"""

    patch = "\n".join(
        difflib.unified_diff(
            original.splitlines(),
            modified.splitlines(),
            fromfile=filename,
            tofile=filename,
            lineterm="",
        )
    )

    return patch


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""\
            Fix macro defines for BOUT++ v4 -> v5

            Please note that this is only slightly better than dumb text replacement. It
            will fix the following:

            * replacement of macros with variables or new names
            * inclusion of correct headers for new variables
            * removal of #if(n)def/#endif blocks that do simple checks for the old
              macro, keeping the appriopriate part, if replaced by a variable
            * change '#if(n)def' for '#if (!)' if the replacment is always defined

            It will try not to replace quoted macro names, but may
            still replace them in strings or comments.

            Please check the diff output carefully!
            """),
    )

    parser.add_argument("files", action="store", nargs="+", help="Input files")
    parser.add_argument(
        "--force", "-f", action="store_true", help="Make changes without asking"
    )
    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Don't print patches"
    )
    parser.add_argument(
        "--patch-only", "-p", action="store_true", help="Print the patches and exit"
    )

    args = parser.parse_args()

    if args.force and args.patch_only:
        raise ValueError("Incompatible options: --force and --patch")

    for filename in args.files:
        with open(filename, "r") as f:
            contents = f.read()
        original = copy.deepcopy(contents)

        modified = apply_fixes(MACRO_REPLACEMENTS, contents)
        patch = create_patch(filename, original, modified)

        if args.patch_only:
            print(patch)
            continue

        if not patch:
            if not args.quiet:
                print("No changes to make to {}".format(filename))
            continue

        if not args.quiet:
            print("\n******************************************")
            print("Changes to {}\n".format(filename))
            print(patch)
            print("\n******************************************")

        if args.force:
            make_change = True
        else:
            make_change = yes_or_no("Make changes to {}?".format(filename))

        if make_change:
            with open(filename, "w") as f:
                f.write(modified)



from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
import numpy as np


def elm_size(dcp, p0, uedge, xmin=None, xmax=None, yind=None, Bbar=None):
    lis = [dcp, p0, uedge]
    if np.size(lis) != 3:
        print("lack of parameters")
        return 0

    if xmin == None:
        xmin = 0
    if xmax == None:
        xmax = 327
    if yind == None:
        yind = 63  # choose the poloidal location for 1D size
    if Bbar == None:
        Bbar = 1.992782  # the normalized magnetic field

    mydcp = dcp
    myp0 = p0
    g = uedge

    PI = 3.1415926
    MU0 = 4.0e-7 * PI

    s = np.shape(mydcp)

    if np.ndim(mydcp) != 3:
        print("dcp should be 3D(t,x,y)")

    nt = s[0]
    nx = s[1]
    ny = s[2]

    Dtheta = g["dy"]  # using correct poloidal angle
    psixy = g["psixy"]
    R = g["Rxy"]
    Bp = g["Bpxy"]
    hthe = g["hthe"]

    Dpsi = np.zeros((nx, ny))
    Dpsi[0, :] = psixy[1, :] - psixy[0, :]
    Dpsi[nx - 1, :] = psixy[nx - 1, :] - psixy[nx - 2, :]
    for i in range(1, nx - 2):
        Dpsi[i, :] = old_div((psixy[i + 1, :] - psixy[i - 1, :]), 2)

    Ddcp1 = np.zeros(nt)
    Ddcp2 = np.zeros(nt)
    Ddcp3 = np.zeros(nt)
    Tp01 = 0.0
    Tp02 = 0.0
    Tp03 = 0.0

    for t in range(nt):
        Ddcp3[t] = (
            2.0
            * PI
            * np.sum(
                mydcp[t, xmin:xmax, :]
                * hthe[xmin:xmax, :]
                * Dtheta[xmin:xmax, :]
                * Dpsi[xmin:xmax, :]
                / Bp[xmin:xmax, :]
            )
        )
        Ddcp2[t] = np.sum(
            mydcp[t, xmin:xmax, :]
            * hthe[xmin:xmax, :]
            * Dtheta[xmin:xmax, :]
            * Dpsi[xmin:xmax, :]
            / (R[xmin:xmax, :] * Bp[xmin:xmax, :])
        )
        Ddcp1[t] = np.sum(
            mydcp[t, xmin:xmax, yind]
            * Dpsi[xmin:xmax, yind]
            / (R[xmin:xmax, yind] * Bp[xmin:xmax, yind])
        )

    Tp03 = (
        2.0
        * PI
        * np.sum(
            myp0[xmin:xmax, :]
            * hthe[xmin:xmax, :]
            * Dtheta[xmin:xmax, :]
            * Dpsi[xmin:xmax, :]
            / Bp[xmin:xmax, :]
        )
    )
    Tp02 = np.sum(
        myp0[xmin:xmax, :]
        * hthe[xmin:xmax, :]
        * Dtheta[xmin:xmax, :]
        * Dpsi[xmin:xmax, :]
        / (R[xmin:xmax, :] * Bp[xmin:xmax, :])
    )
    Tp01 = np.sum(
        myp0[xmin:xmax, yind]
        * Dpsi[xmin:xmax, yind]
        / (R[xmin:xmax, yind] * Bp[xmin:xmax, yind])
    )

    s1 = np.zeros(nt)
    s2 = np.zeros(nt)
    s3 = np.zeros(nt)
    E_loss = np.zeros(nt)

    s1 = old_div(-Ddcp1, Tp01)  # 1D elm size
    s2 = old_div(-Ddcp2, Tp02)  # 2D elm size
    s3 = old_div(-Ddcp3, Tp03)  # 3D elm size

    E_loss = -Ddcp3 * (0.5 * Bbar * Bbar / MU0)  # energy loss, unit J
    E_total = Tp03 * (0.5 * Bbar * Bbar / MU0)  # total energy, unit J

    class ELM:
        pass

    elmsize = ELM()
    elmsize.s1 = s1
    elmsize.s2 = s2
    elmsize.s3 = s3
    elmsize.E_loss = E_loss
    elmsize.E_total = E_total

    return elmsize



from __future__ import print_function
######
# Computes the rms of a variable and prints the growth rate value at the last timestep
######

import numpy as np
from boutdata.collect import collect
from boututils.moment_xyzt import moment_xyzt


path = "./data/"


p = collect("P", path=path)
rmsp_f = moment_xyzt(p[:, 34:35, 32:33, :], "RMS").rms


print(np.gradient(np.log(rmsp_f[:, 0, 0]))[-1])



#!/usr/bin/env python3

from boutdata.data import BoutOutputs
from matplotlib import pyplot
from sys import exit

Ntests = 4

d = BoutOutputs("data")
# d = BoutOutputs("data", caching=True, info=False)

for i in range(1, Ntests + 1):
    pyplot.figure(i)
    pyplot.subplot(231)
    pyplot.pcolor(d["f" + str(i)][1:-1, 0, :].T)
    pyplot.colorbar()
    pyplot.title("f" + str(i))
    pyplot.xlabel("X")
    pyplot.ylabel("Z")
    pyplot.subplot(232)
    pyplot.pcolor(d["sol" + str(i)][1:-1, 0, :].T)
    pyplot.colorbar()
    pyplot.title("sol" + str(i))
    pyplot.xlabel("X")
    pyplot.ylabel("Z")
    pyplot.subplot(233)
    pyplot.pcolor(d["absolute_error" + str(i)][1:-1, 0, :].T)
    pyplot.colorbar()
    pyplot.title("absolute_error" + str(i))
    pyplot.xlabel("X")
    pyplot.ylabel("Z")
    pyplot.subplot(234)
    pyplot.pcolor(d["b" + str(i)][2:-2, 0, :].T)
    pyplot.colorbar()
    pyplot.title("b" + str(i))
    pyplot.xlabel("X")
    pyplot.ylabel("Z")
    pyplot.subplot(235)
    pyplot.pcolor(d["bcheck" + str(i)][2:-2, 0, :].T)
    pyplot.colorbar()
    pyplot.title("bcheck" + str(i))
    pyplot.xlabel("X")
    pyplot.ylabel("Z")
    pyplot.subplot(236)
    pyplot.pcolor(d["b" + str(i)][2:-2, 0, :].T - d["bcheck" + str(i)][2:-2, 0, :].T)
    pyplot.colorbar()
    pyplot.title("b" + str(i) + "-bcheck" + str(i))
    pyplot.xlabel("X")
    pyplot.ylabel("Z")

pyplot.show()

exit(0)



#!/usr/bin/env python3
from __future__ import division
from past.utils import old_div

import matplotlib.pyplot as plt
import numpy as np
from boututils.moment_xyzt import moment_xyzt
from boututils.datafile import DataFile
from boutdata.collect import collect
import os
from pathlib import Path

# Dynamic matplotlib settings
from matplotlib import rcParams

rcParams["font.size"] = 20.0
rcParams["legend.fontsize"] = "small"
rcParams["lines.linewidth"] = 2

if not os.path.exists("image"):
    os.makedirs("image")
filename = Path(__file__).with_name("cbm18_dens8.grid_nx68ny64.nc")
with DataFile(str(filename)) as f:
    g = {v: f.read(v) for v in f.keys()}

psi = old_div((g["psixy"][:, 32] - g["psi_axis"]), (g["psi_bndry"] - g["psi_axis"]))

path = "./data"

plt.figure()

p0 = collect("P0", path=path)

p = collect("P", path=path)
res = moment_xyzt(p, "RMS", "DC")
rmsp = res.rms
dcp = res.dc
nt = dcp.shape[0]

plt.plot(psi, p0[:, 32], "k--", label="t=0")
plt.plot(psi, p0[:, 32] + dcp[nt // 4, :, 32], "r-", label="t=" + np.str(nt // 4))
plt.plot(psi, p0[:, 32] + dcp[nt // 2, :, 32], "g-", label="t=" + np.str(nt // 2))
plt.plot(
    psi, p0[:, 32] + dcp[3 * nt // 4, :, 32], "b-", label="t=" + np.str(3 * nt // 4)
)
plt.plot(psi, p0[:, 32] + dcp[-1, :, 32], "c-", label="t=" + np.str(nt))

plt.legend()
# plt.xlim(0.6, 1.0)
plt.xlabel(r"Normalized poloidal flux ($\psi$)")
plt.ylabel(r"$\langle p\rangle_\xi$")
plt.title(r"Pressure")
xmin, xmax = plt.xlim()
ymin, ymax = plt.ylim()


# plt.savefig('image/plotcollapse.png', bbox_inches='tight')
# plt.savefig('image/plotcollapse.eps', bbox_inches='tight')


plt.tight_layout()

print("Showing plot")
plt.show()



# MPark.Variant
#
# Copyright Michael Park, 2015-2017
#
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE.md or copy at http://boost.org/LICENSE_1_0.txt)

import os
import pprint
import subprocess

result = {}

std_flags = os.getenv('STDFLAGS')
for std_flag in std_flags.split() if std_flags is not None else ['']:
  os.environ['CXXFLAGS'] = std_flag
  for exceptions in ['OFF', 'ON']:
    config = '{}-{}'.format(filter(str.isalnum, std_flag), exceptions)
    build_dir = 'build-{}'.format(config)
    os.mkdir(build_dir)
    os.chdir(build_dir)
    result[config] = {
        'Configure': None,
        'Build-Debug': None,
        'Build-Release': None,
        'Test-Debug': None,
        'Test-Release': None
    }

    tests = os.environ['TESTS'].split()
    if std_flag.endswith(('11', '1y', '14', '1z')) and 'libc++' in tests:
      tests.remove('libc++')

    configure = [
      'cmake', '-G', os.environ['GENERATOR'],
               '-DMPARK_VARIANT_EXCEPTIONS={}'.format(exceptions),
               '-DMPARK_VARIANT_INCLUDE_TESTS={}'.format(';'.join(tests)),
               '..',
    ]
    result[config]['Configure'] = subprocess.call(configure)
    if result[config]['Configure'] == 0:
      for build_type in ['Debug', 'Release']:
        result[config]['Build-{}'.format(build_type)] = subprocess.call([
          'cmake', '--build', '.', '--config', build_type])
        if result[config]['Build-{}'.format(build_type)] == 0:
          result[config]['Test-{}'.format(build_type)] = subprocess.call([
            'ctest', '--output-on-failure', '--build-config', build_type])
    os.chdir('..')

pprint.pprint(result)
exit(any(status != 0 for d in result.itervalues() for status in d.itervalues()))



from __future__ import print_function
from boututils.datafile import DataFile
from boututils.volume_integral import volume_integral
from boutdata.collect import collect

# Integrate over a volume


path = "./data/"

gfile = "./cbm18_dens8.grid_nx68ny64.nc"

with DataFile(gfile) as f:
    g = {v: f.read(v) for v in f.keys()}

var = collect("P", path=path)

sol = volume_integral(var, g)

print(sol)



# -*- coding: utf-8 -*-
"""
Created on Wed Oct  5 15:55:04 2016

@author: yolen
"""

import boutdata
import matplotlib.pyplot as plt

path = "."
E_n_ = boutdata.collect("E_N", xguards=False, yguards=False, path=path)
# S_n_ = boutdata.collect('S_N',xguards=False, yguards = False, path = path)
n = boutdata.collect("N", xguards=False, path=path)

print("shape E_n = {}".format(E_n_.shape))
# n_source_ = boutdata.collect('n_source_',xguards=False)

plt.subplot(2, 2, 1)
plt.plot(E_n_[1, 0, 0, :], "ro")
plt.title("error")
# plt.colorbar()
plt.subplot(2, 2, 2)
plt.plot(n[1, 0, 0, :], "ro")
plt.title("solution")
# plt.colorbar()
# plt.subplot(2,2,3)
# plt.plot(S_n_[1,1,0,:],'ro')
# plt.title('source')
# plt.colorbar()

plt.show()



#!/usr/bin/env python3

from boutdata import collect
from boututils.datafile import DataFile
from matplotlib import pyplot
import numpy
from sys import exit

# Set default colormap
pyplot.rcParams["image.cmap"] = "viridis"
# pyplot.rcParams['image.cmap'] = 'plasma'

x = collect("x", path="data", xguards=False, yguards=False)
rhs = collect("rhs", path="data", xguards=False, yguards=False)
check = collect("check", path="data", xguards=False, yguards=False)
x2 = collect("x2", path="data", xguards=False, yguards=False)
rhs2 = collect("rhs2", path="data", xguards=False, yguards=False)
check2 = collect("check2", path="data", xguards=False, yguards=False)

grid = DataFile("data/d3d_119919.nc")
R = grid["Rxy"][2:-2, :]
Z = grid["Zxy"][2:-2, :]

pyplot.subplot(231)
pyplot.pcolor(R, Z, x)
pyplot.colorbar()
pyplot.title("x")

pyplot.subplot(232)
pyplot.pcolor(R, Z, rhs)
pyplot.colorbar()
pyplot.title("rhs")

pyplot.subplot(233)
pyplot.pcolor(R, Z, check)
pyplot.colorbar()
pyplot.title("check")

pyplot.subplot(234)
pyplot.pcolor(R, Z, x2)
pyplot.colorbar()
pyplot.title("x2")

pyplot.subplot(235)
pyplot.pcolor(R, Z, rhs2)
pyplot.colorbar()
pyplot.title("rhs2")

pyplot.subplot(236)
pyplot.pcolor(R, Z, check2)
pyplot.colorbar()
pyplot.title("check2")

pyplot.show()

exit(0)



from __future__ import division
from builtins import range
from past.utils import old_div
import numpy as np
import matplotlib.pyplot as plt
from boututils.datafile import DataFile
from matplotlib.ticker import (
    FixedFormatter,
    FormatStrFormatter,
    AutoLocator,
    AutoMinorLocator,
)

with DataFile("./cbm18_dens8.grid_nx68ny64.nc") as f:
    g = {v: f.read(v) for v in f.keys()}

majorLocator = AutoLocator()
majorFormatter = FormatStrFormatter("%3.0e")
minorLocator = AutoMinorLocator()
Fm = FixedFormatter(
    [
        "0",
        "$1 \\times 10^4$",
        "$2 \\times  10^4$",
        "$3 \\times 10^4$",
        "$4 \\times 10^4$",
    ]
)
Fm2 = FixedFormatter(["0", "$2 \\times 10^5$", "$4 \\times  10^5$", "$6 \\times 10^5$"])

bxy = g.get("Bxy")
p = g.get("pressure")
jpar0 = g.get("Jpar0")
psixy = g.get("psixy")
btxy = g.get("Btxy")
shiftangle = g.get("ShiftAngle")

nx = g.get("nx")
ny = g.get("ny")

q = np.zeros((nx, ny))
for i in range(ny):
    q[:, i] = old_div(-shiftangle, (2 * np.pi))


xarr = psixy[:, 0]
xarr = old_div((xarr + 0.854856), (0.854856 + 0.0760856))


fig = plt.figure()
plt.plot(xarr, q[:, 32])
plt.xlabel("normalized $\psi$", fontsize=25)
plt.ylabel("$q$", rotation="horizontal", fontsize=25)

fig.set_tight_layout(True)

fig, ax1 = plt.subplots()

ax1.plot(xarr, p[:, 32], "r-", markevery=1, linewidth=3)
ax1.set_xlabel("normalized $\psi$", fontsize=25)
# Make the y-axis label and tick labels match the line color.
ax1.set_ylabel("Pressure [Pa]", color="k", fontsize=25)

# set y limit
ax1.set_ylim(0, 40000, 10000)

# define ticks#
ax1.yaxis.set_ticks(np.arange(0, 40000, 10000))
# ax1.yaxis.set_major_locator(majorLocator)
# ax1.yaxis.set_major_formatter(majorFormatter)
ax1.yaxis.set_major_formatter(Fm)

# for the minor ticks, use no labels; default NullFormatter
ax1.xaxis.set_minor_locator(AutoMinorLocator())
ax1.yaxis.set_minor_locator(AutoMinorLocator(10))

# format tick labels
for tl in ax1.get_yticklabels():
    tl.set_color("k")


ax2 = ax1.twinx()

s2 = -jpar0

ax2.plot(xarr, s2[:, 32], "r-", markevery=1, linewidth=3)
ax2.set_ylabel("$J_\parallel  [A/m^2]$", color="k", fontsize=25)
ax2.set_ylim(0, 600000)
ax2.yaxis.set_ticks(np.arange(0, 600000, 200000))
ax2.yaxis.set_major_formatter(Fm2)
for tl in ax2.get_yticklabels():
    tl.set_color("k")


fig.set_tight_layout(True)

plt.show()


# plt.savefig('2d.png', transparent=True)



#!/usr/bin/env python3

from boutdata import collect
from matplotlib import pyplot
from sys import argv, exit

datadir = argv[1]

yind = int(argv[2])

xg = 2

f = collect("f", path=datadir)[xg:-xg, :, :]
rhs = collect("rhs", path=datadir)[xg:-xg, :, :]
rhs_check = collect("rhs_check", path=datadir)[xg:-xg, :, :]
error = collect("error", path=datadir)[xg:-xg, :, :]

pyplot.subplot(221)
pyplot.pcolormesh(f[:, yind, :])
pyplot.colorbar()
pyplot.title("f")
pyplot.subplot(222)
pyplot.pcolormesh(rhs[:, yind, :])
pyplot.colorbar()
pyplot.title("rhs")
pyplot.subplot(223)
pyplot.pcolormesh(rhs_check[:, yind, :])
pyplot.colorbar()
pyplot.title("rhs_check")
pyplot.subplot(224)
pyplot.pcolormesh(error[:, yind, :])
pyplot.colorbar()
pyplot.title("error")

pyplot.show()

exit(0)



#!/usr/bin/env python

from __future__ import print_function
from builtins import str
from boututils.run_wrapper import shell, launch
from boutdata.collect import collect

import numpy as np

print("Making 3-field ELM simulation")
shell("make > make.log")

# Run simulation
nproc = 2

s, out = launch("./elm_pb ", nproc=nproc, pipe=True)

with open("run.log", "w") as f:
    f.write(out)

# Get time base
tarr = collect("t_array", path="data")
nt = len(tarr)

# Read pressure
p = collect("P", path="data")

# Calculate RMS in toroidal direction
prms = np.sqrt(np.mean(p**2, axis=3))

growth = np.gradient(np.log(prms[:, 42, 32]))

# Final growth-rate
gamma = growth[-2]

import matplotlib.pyplot as plt

plt.plot(tarr, prms[:, 42, 32], label="Outboard midplane")
plt.plot(
    [tarr[0], tarr[-1]],
    [prms[-1, 42, 32] * np.exp(gamma * (tarr[0] - tarr[-1])), prms[-1, 42, 32]],
    "--",
    label=r"$\gamma =$" + str(gamma),
)

plt.yscale("log")
plt.grid()

plt.xlabel(r"Time [$1/\tau_A$]")
plt.ylabel("Pressure RMS")

plt.legend(loc="upper left")

plt.savefig("growth.pdf")

plt.show()

############### Poloidal plot

plt.figure()

# Take a poloidal slice at fixed toroidal angle
from boutdata.pol_slice import pol_slice

p2d = pol_slice(p[-1, :, :, :], "cbm18_dens8.grid_nx68ny64.nc", n=15, zangle=0.0)

# Read grid file to get coordinates
from boututils.datafile import DataFile

g = DataFile("cbm18_dens8.grid_nx68ny64.nc")

Rxy = g.read("Rxy")  # Major radius [m]
Zxy = g.read("Zxy")  # Height [m]

plt.contourf(Rxy, Zxy, p2d, 30)
plt.axis("equal")  # Maintain aspect ratio

plt.colorbar()  # Plot a bar down the side with a color scale

plt.savefig("poloidal_slice.pdf")

plt.show()



import os
from tempfile import NamedTemporaryFile as tmpf
from stencils_sympy import dirichlet, neumann, simp, Symbol, Matrix, ccode


def gen_code(order, matrix_type):
    x = [Symbol("spacing%d" % i) for i in range(order)]
    matrix = matrix_type(x)
    A = Matrix(order, order, matrix)

    try:
        iA = A.inv()
    except:
        import sys

        print(A, matrix, file=sys.stderr)
        raise
    return ccode(simp(sum([iA[0, i] * Symbol("value%d" % i) for i in range(order)])))


def run(cmd):
    print(cmd)
    out = os.system(cmd)
    assert out == 0


if __name__ == "__main__":
    with tmpf("w", dir=".", delete=False) as f:
        f.write("namespace {\n")
        f.write(
            """
inline BoutReal pow(BoutReal val, int exp) {
  //constexpr int expval = exp;
  //static_assert(expval == 2 or expval == 3, "This pow is only for exponent 2 or 3");
  if (exp == 2) {
    return val * val;
  }
  ASSERT3(exp == 3);
  return val * val * val;
}
"""
        )

        for order in range(1, 4):
            for matrix in dirichlet, neumann:
                if order == 1 and matrix == neumann:
                    continue
                print(f"generating {matrix.name}_o{order}")
                args = ", ".join(
                    [
                        "BoutReal spacing%d, BoutReal value%d" % (i, i)
                        for i in range(order)
                    ]
                )
                f.write(
                    f"inline BoutReal stencil_{matrix.name}_o{order}({args}) {{\n  return "
                )
                f.write(gen_code(order, matrix))
                f.write(";\n}\n")
        f.write("}\n")
    run("clang-format -i " + f.name)
    run(f"mv {f.name} {__file__[:-3]}")



# Plots an analysis of the linear growth rate
#
# Input argument is the directory containing data files.
#
# Example:
#     $ python plot_linear.py data/
#

from boutdata import collect
import numpy as np
import matplotlib.pyplot as plt
import os
import sys

if len(sys.argv) != 2:
    raise ValueError(f"Usage: {sys.argv[0]} path")

# Path to the data
path = sys.argv[1]

# Read pressure at last time point, to find peak
p = collect("P", path=path, tind=-1).squeeze()
prms = np.sqrt(np.mean(p**2, axis=-1))

pyprof = np.amax(prms, axis=0)
yind = np.argmax(pyprof)
pxprof = prms[:, yind]
xind = np.argmax(pxprof)
print(f"Peak amplitude at x = {xind}, y = {yind}")

# Read pressure time history at index of peak amplitude
p = collect("P", path=path, xind=xind, yind=yind).squeeze()

# p = p[:,:-1] # Remove point in Z

prms = np.sqrt(np.mean(p**2, axis=-1))

t = collect("t_array", path=path)
dt = t[1] - t[0]

gamma = np.gradient(np.log(prms)) / dt
growth_rate = np.mean(gamma[len(gamma) // 2 :])
growth_rate_std = np.std(gamma[len(gamma) // 2 :])

print(f"Mean growth rate: {growth_rate} +/- {growth_rate_std}")

fig, axs = plt.subplots(2, 2)

ax = axs[0, 0]
ax.plot(pyprof)
ax.set_xlabel("Y index")
ax.set_ylabel("RMS pressure")
ax.axvline(yind, linestyle="--", color="k")
ax.text(yind, 0.5 * pyprof[yind], f"y = {yind}")

ax = axs[0, 1]
ax.plot(pxprof)
ax.set_xlabel("X index")
ax.set_ylabel("RMS pressure")
ax.axvline(xind, linestyle="--", color="k")
ax.text(xind, 0.5 * pxprof[xind], f"x = {xind}")

ax = axs[1, 0]
ax.plot(t, prms)
ax.set_xlabel(r"Time [$\tau_A$]")
ax.set_ylabel("RMS pressure")
ax.set_yscale("log")
ax.plot(t, prms[-1] * np.exp(growth_rate * (t - t[-1])), "--k")

ax = axs[1, 1]
ax.plot(t, gamma)
ax.set_xlabel(r"Time [$\tau_A$]")
ax.set_ylabel("Growth rate [$\omega_A$]")
ax.axhline(growth_rate, linestyle="--", color="k")
ax.text(
    (t[-1] + t[0]) / 4,
    growth_rate * 1.2,
    rf"$\gamma = {growth_rate:.3f}\pm {growth_rate_std:.3f} \omega_A$",
)

plt.savefig(os.path.join(path, "linear_growth.pdf"))
plt.savefig(os.path.join(path, "linear_growth.png"))

plt.tight_layout()
plt.show()



#!/usr/bin/env python3
# /// script
# requires-python = ">=3.11"
# dependencies = [
#     "ruamel-yaml",
#     "unidecode",
# ]
# ///
import argparse
import subprocess
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

from ruamel.yaml import YAML
from unidecode import unidecode


@dataclass
class KnownAuthor:
    family_names: str
    given_names: str


TOP_DIR = Path(__file__).parent.parent
CITATION_FILE = TOP_DIR / "CITATION.cff"
NONHUMAN_AUTHORS = ("github", "dependabot")
KNOWN_AUTHORS = {
    "bendudson": KnownAuthor("Dudson", "Benjamin"),
    "brey": KnownAuthor("Breyiannis", "George"),
    "David Schwörer": KnownAuthor("Bold", "David"),
    "dschwoerer": KnownAuthor("Bold", "David"),
    "hahahasan": KnownAuthor("Muhammed", "Hasan"),
    "Ilon Joseph - x31405": KnownAuthor("Joseph", "Ilon"),
    "kangkabseok": KnownAuthor("Kang", "Kab Seok"),
    "loeiten": KnownAuthor("Løiten", "Michael"),
    "Michael Loiten Magnussen": KnownAuthor("Løiten", "Michael"),
    "Maxim Umansky - x26041": KnownAuthor("Umansky", "Maxim"),
    "nick-walkden": KnownAuthor("Walkden", "Nicholas"),
    "ZedThree": KnownAuthor("Hill", "Peter"),
    "tomc271": KnownAuthor("Chapman", "Tom"),
    "j-b-o": KnownAuthor("Bold", "Jessica"),
    "BS": KnownAuthor("Brendan", "Shanahan"),
}


def get_authors_from_git():
    output = subprocess.run(
        ["git", "log", "--format='%aN %aE'"],
        capture_output=True,
        check=True,
        text=True,
    )

    authors_string = output.stdout
    authors_list = authors_string.split("\n")
    authors_without_quotes = [a.strip("'") for a in authors_list]

    distinct_authors = set(authors_without_quotes)
    distinct_authors_list_without_empty_strings = [
        a.rsplit(maxsplit=1) for a in distinct_authors if a
    ]
    authors_with_emails = defaultdict(list)
    for author, email in distinct_authors_list_without_empty_strings:
        authors_with_emails[author].append(email)

    return authors_with_emails


def parse_cff_file():
    yaml = YAML(typ="rt")
    with open(CITATION_FILE, "r", encoding="UTF-8") as stream:
        return yaml.load(stream)


def get_authors_from_cff_file():
    file_contents = parse_cff_file()
    try:
        return file_contents["authors"]
    except KeyError as key_error:
        raise ValueError(f"Failed to find section:{key_error} in {CITATION_FILE}")


class ExistingAuthorNames:
    def __init__(self, existing_authors):
        self._existing_author_names = [
            (unidecode(a.get("given-names")), unidecode(a.get("family-names")))
            for a in existing_authors
        ]

    def last_name_matches_surname_and_first_name_or_first_letter_matches_given_name(
        self, last_name, first_name
    ):
        # Last name matches surname
        matches = [
            n
            for n in self._existing_author_names
            if n[1].casefold() == last_name.casefold()
        ]

        for match in matches:
            # The given name also matches author first name
            if match[0].casefold() == first_name.casefold():
                return True
            # The first initial matches author first name
            if match[0][0].casefold() == first_name[0].casefold():
                return True

    def first_name_matches_surname_and_last_name_matches_given_name(
        self, first_name, last_name
    ):
        # First name matches surname
        matches = [
            n
            for n in self._existing_author_names
            if n[1].casefold() == first_name.casefold()
        ]

        # The given name also matches author last name
        for match in matches:
            if match[0].casefold() == last_name.casefold():
                return True

    def surname_matches_whole_author_name(self, author):
        surname_matches = [
            n
            for n in self._existing_author_names
            if n[1].casefold() == author.casefold()
        ]
        if len(surname_matches) > 0:
            return True

    def given_name_matches_matches_whole_author_name(self, author):
        given_name_matches = [
            n
            for n in self._existing_author_names
            if n[0].casefold() == author.casefold()
        ]
        if len(given_name_matches) > 0:
            return True

    def combined_name_matches_whole_author_name(self, author):
        combined_name_matches = [
            n
            for n in self._existing_author_names
            if (n[0] + n[1]).casefold() == author.casefold()
        ]
        if len(combined_name_matches) > 0:
            return True

    def combined_name_reversed_matches(self, author):
        combined_name_reversed_matches = [
            n
            for n in self._existing_author_names
            if (n[1] + n[0]).casefold() == author.casefold()
        ]
        if len(combined_name_reversed_matches) > 0:
            return True

    def author_name_is_first_initial_and_surname_concatenated(self, author):
        first_character = author[0]
        remaining_characters = author[1:]
        # Second part of name matches surname
        matches = [
            n
            for n in self._existing_author_names
            if n[1].casefold() == remaining_characters.casefold()
        ]
        for match in matches:
            # The first initial matches author first name
            if match[0][0].casefold() == first_character.casefold():
                return True


def author_found_in_existing_authors(author, existing_authors):
    existing_author_names = ExistingAuthorNames(existing_authors)

    names = author.split()
    first_name = unidecode(names[0].replace(",", ""))
    last_name = unidecode(names[-1])

    if existing_author_names.last_name_matches_surname_and_first_name_or_first_letter_matches_given_name(
        last_name, first_name
    ):
        return True

    if existing_author_names.first_name_matches_surname_and_last_name_matches_given_name(
        first_name, last_name
    ):
        return True

    if existing_author_names.surname_matches_whole_author_name(author):
        return True

    if existing_author_names.given_name_matches_matches_whole_author_name(author):
        return True

    if existing_author_names.combined_name_matches_whole_author_name(author):
        return True

    if existing_author_names.combined_name_reversed_matches(author):
        return True

    if existing_author_names.author_name_is_first_initial_and_surname_concatenated(
        author
    ):
        return True

    return False


def is_nonhuman(author: str) -> bool:
    lower_case_author = author.casefold()
    return any(nonhuman in lower_case_author for nonhuman in NONHUMAN_AUTHORS)


def update_citations():
    authors_from_git = get_authors_from_git()
    existing_authors = get_authors_from_cff_file()

    known_authors = [a for a in authors_from_git if a in KNOWN_AUTHORS]
    human_authors = {a: e for a, e in authors_from_git.items() if not is_nonhuman(a)}
    authors_to_search_for = {
        a: e for a, e in human_authors.items() if a not in known_authors
    }

    unrecognised_authors = {
        a: e
        for a, e in authors_to_search_for.items()
        if not author_found_in_existing_authors(a, existing_authors)
    }

    if not unrecognised_authors:
        return

    for author, email in unrecognised_authors.items():
        print(f"{author} (email(s): {email})")

    while True:
        reply = (
            input("\nThe above authors were not recognised. Add to citations? [y/N] ")
            .lower()
            .strip()
        )
        if not reply or reply[0] == "n":
            return
        if reply[0] == "y":
            break

    new_authors = []
    for author in unrecognised_authors:
        if " " not in author:
            # This is just a GitHub handle, skip
            continue
        first_name, last_name = author.rsplit(maxsplit=1)
        new_authors.append({"family-names": last_name, "given-names": first_name})

    print(f"Adding new authors:\n{new_authors}")

    yaml_file = parse_cff_file()
    yaml_file["authors"].extend(new_authors)

    # Try to preserve indentation and quotes as much as possible
    yaml = YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.preserve_quotes = True
    with open(CITATION_FILE, "w", encoding="UTF-8") as f:
        yaml.dump(yaml_file, f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Update CITATIONS.cff based on git authors"
    )
    parser.parse_args()

    update_citations()



#!/usr/bin/env python3

import argparse
import re
import fileinput
import sys

nonmembers = {
    "DC": ["DC", 1],
    "slice": ["sliceXZ", 2],
}

coordinates = [
    "outputVars",
    "dx",
    "dy",
    "dz",
    "zlength",
    "non_uniform",
    "d1_dx",
    "d1_dy",
    "J",
    "Bxy",
    "g11",
    "g22",
    "g33",
    "g12",
    "g13",
    "g23",
    "g_11",
    "g_22",
    "g_33",
    "g_12",
    "g_13",
    "g_23",
    "G1_11",
    "G1_22",
    "G1_33",
    "G1_12",
    "G1_13",
    "G2_11",
    "G2_22",
    "G2_33",
    "G2_12",
    "G2_23",
    "G3_11",
    "G3_22",
    "G3_33",
    "G3_13",
    "G3_23",
    "G1",
    "G2",
    "G3",
    "ShiftTorsion",
    "IntShiftTorsion",
    "geometry",
    "calcCovariant",
    "calcContravariant",
    "jacobian",
]

local_mesh = [
    ("ngx", "LocalNx"),
    ("ngy", "LocalNy"),
]

warnings = [
    (r"\^", "Use pow(a,b) instead of a^b"),
    (r"\.max\(", "Use max(a) instead of a.max()"),
    (
        "ngz",
        (
            "ngz is changed to LocalNz in v4."
            " The extra point in z has been removed."
            " Change ngz -> LocalNz, and ensure that"
            " the number of points are correct"
        ),
    ),
]


def fix_nonmembers(line_text, filename, line_num, replace=False):
    """Replace member functions with nonmembers"""

    old_line_text = line_text

    for old, (new, num_args) in nonmembers.items():
        pattern = re.compile("(\w*)\.{}\(".format(old))
        matches = re.findall(pattern, line_text)
        for match in matches:
            replacement = "{func}({object}".format(func=new, object=match)
            if num_args > 1:
                replacement += ", "
            line_text = re.sub(pattern, replacement, line_text)
            if not replace:
                name_num = "{name}:{num}:".format(name=filename, num=line_num)
                print(
                    "{name_num}{line}".format(name_num=name_num, line=old_line_text),
                    end="",
                )
                print(" " * len(name_num) + line_text)
    if replace:
        return line_text


def fix_subscripts(line_text, filename, line_num, replace=False):
    """Replace triple square brackets with round brackets

    Should also check that the variable is a Field3D/Field2D - but doesn't
    """

    old_line_text = line_text
    # Catch both 2D and 3D arrays
    pattern = re.compile(r"\[([^[]*)\]\[([^[]*)\](?:\[([^[]*)\])?")
    matches = re.findall(pattern, line_text)
    for match in matches:
        # If the last group is non-empty, then it was a 3D array
        if len(match[2]):
            replacement = r"(\1, \2, \3)"
        else:
            replacement = r"(\1, \2)"
        line_text = re.sub(pattern, replacement, line_text)
        if not replace:
            name_num = "{name}:{num}:".format(name=filename, num=line_num)
            print(
                "{name_num}{line}".format(name_num=name_num, line=old_line_text), end=""
            )
            print(" " * len(name_num) + line_text)
    if replace:
        return line_text


def fix_coordinates(line_text, filename, line_num, replace=False):
    """Fix variables that have moved from mesh to coordinates"""

    old_line_text = line_text

    for var in coordinates:
        pattern = re.compile("mesh->{}".format(var))
        matches = re.findall(pattern, line_text)
        for match in matches:
            line_text = re.sub(
                pattern, "mesh->coordinates()->{}".format(var), line_text
            )
            if not replace:
                name_num = "{name}:{num}:".format(name=filename, num=line_num)
                print(
                    "{name_num}{line}".format(name_num=name_num, line=old_line_text),
                    end="",
                )
                print(" " * len(name_num) + line_text)
    if replace:
        return line_text


def fix_local_mesh_size(line_text, filename, line_num, replace=False):
    """Replaces ng@ with LocalNg@, where @ is in {x,y,z}"""

    old_line_text = line_text

    for lm in local_mesh:
        pattern = re.compile(lm[0])
        matches = re.findall(pattern, line_text)
        for match in matches:
            line_text = re.sub(pattern, lm[1], line_text)
            if not replace:
                name_num = "{name}:{num}:".format(name=filename, num=line_num)
                print(
                    "{name_num}{line}".format(name_num=name_num, line=old_line_text),
                    end="",
                )
                print(" " * len(name_num) + line_text)
    if replace:
        return line_text


def throw_warnings(line_text, filename, line_num):
    """Throws a warning for ^, .max() and ngz"""

    for warn in warnings:
        pattern = re.compile(warn[0])
        matches = re.findall(pattern, line_text)
        for match in matches:
            name_num = "{name}:{num}:".format(name=filename, num=line_num)
            # stdout is redirected to the file if --replace is given,
            # therefore use stderr
            sys.stderr.write(
                "{name_num}{line}".format(name_num=name_num, line=line_text)
            )
            # Coloring with \033[91m, end coloring with \033[0m\n
            sys.stderr.write(
                " " * len(name_num)
                + "\033[91m!!!WARNING: {}\033[0m\n\n".format(warn[1])
            )


if __name__ == "__main__":
    epilog = """
    Currently bout_3to4 can detect the following transformations are needed:
        - Triple square brackets instead of round brackets for subscripts
        - Field member functions that are now non-members
        - Variables/functions that have moved from Mesh to Coordinates

    Note that in the latter case of transformations, you will still need to manually add
        Coordinates *coords = mesh->coordinates();
    to the correct scopes
    """

    parser = argparse.ArgumentParser(
        description="A little helper for upgrading from BOUT++ version 3 to version 4",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=epilog,
    )
    parser.add_argument(
        "-r", "--replace", action="store_true", help="Actually make the fix"
    )
    parser.add_argument("files", nargs="+", help="Files to process")

    args = parser.parse_args()

    # Loops over all lines across all files
    for line in fileinput.input(files=args.files, inplace=args.replace):
        filename = fileinput.filename()
        line_num = fileinput.filelineno()

        # Apply the transformations and then update the line if we're doing a replacement
        new_line = fix_nonmembers(line, filename, line_num, args.replace)
        line = new_line if args.replace else line

        new_line = fix_subscripts(line, filename, line_num, args.replace)
        line = new_line if args.replace else line

        new_line = fix_coordinates(line, filename, line_num, args.replace)
        line = new_line if args.replace else line

        new_line = fix_local_mesh_size(line, filename, line_num, args.replace)
        line = new_line if args.replace else line

        new_line = throw_warnings(line, filename, line_num)

        # If we're doing a replacement, then we need to print all lines, without a newline
        if args.replace:
            print(line, end="")



def _resolve_slice(ind, num):
    if isinstance(ind, slice):
        ret = list(ind.indices(num))
        d = ret[1] - ret[0]
        rem = d % ret[2]
        if rem != 0:
            ret[1] -= rem
            ret[1] += ret[2]
        if d * ret[2] < 0:
            return [0, 0, 1]
        return ret
    if ind < 0:
        ind = num + ind
    if ind < 0 or ind >= num:
        raise IndexError("%d is out of range [%d:%d]" % (ind, 0, num))
    return [ind, ind + 1, 1]


def _resolve_slices(inds, nums):
    if len(inds) != len(nums):
        raise IndexError(
            "Expexted %d dimensions, but got %d dimensions." % (len(nums), len(inds))
        )
    ret = []
    for i in range(len(inds)):
        ret.append(_resolve_slice(inds[i], nums[i]))
    return ret



from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div

###
# computes average growth rate for all points at the final timestep
# computes average growth rate for points in the mead plane at the final timestep
###
import numpy as np
from boutdata.collect import collect
from boututils.moment_xyzt import moment_xyzt

path = "./data/"

p = collect("P", path=path)

nmpy = old_div(p.shape[2], 2)  # define mead plane

ik = 50  # disregard the first ik timesteps


def gr(p):
    rmsp_f = moment_xyzt(p, "RMS").rms

    ni = np.shape(rmsp_f)[1]
    nj = np.shape(rmsp_f)[2]

    growth = np.zeros((ni, nj))

    for i in range(ni):
        for j in range(nj):
            growth[i, j] = np.gradient(np.log(rmsp_f[ik::, i, j]))[-1]

    return growth


growth = gr(p)

d = np.ma.masked_array(growth, np.isnan(growth))

# masked arrays
# http://stackoverflow.com/questions/5480694/numpy-calculate-averages-with-nans-removed

print("Total mean value = ", np.mean(np.ma.masked_array(d, np.isinf(d))))
mm = np.ma.masked_array(growth[:, nmpy], np.isnan(growth[:, nmpy]))
if np.isinf(np.mean(mm)):
    print("There is an Inf value in the mead plane")
    print(
        "Mean value of floating numbers in mead plane is = ",
        np.mean(np.ma.masked_array(mm, np.isinf(mm))),
    )
else:
    print("Mean value in mead plane= ", np.mean(mm))



from __future__ import division
from past.utils import old_div

from numpy import save, load, angle
import matplotlib.pyplot as plt

fphi = load("fphi.npy")

fte = load("fte.npy")
phase_te = angle(old_div(fphi, fte))
save("phase_te", phase_te)
plt.figure()
plt.plot(mean(mean(phase_te[:, :, 3, :], axis=1), axis=1))
plt.title("Te")
plt.savefig("image/phase_te.png")
plt.savefig("image/phase_te.eps")

fti = load("fti.npy")
phase_ti = angle(old_div(fphi, fti))
save("phase_ti", phase_ti)
plt.figure()
plt.plot(mean(mean(phase_ti[:, :, 3, :], axis=1), axis=1))
plt.title("ti")
plt.savefig("image/phase_ti.png")
plt.savefig("image/phase_ti.eps")

fni = load("fni.npy")
phase_ni = angle(old_div(fphi, fni))
save("phase_ni", phase_ni)
plt.figure()
plt.plot(mean(mean(phase_ni[:, :, 3, :], axis=1), axis=1))
plt.title("ni")
plt.savefig("image/phase_ni.png")
plt.savefig("image/phase_ni.eps")

plt.show()



from __future__ import division
from __future__ import print_function

from boutdata.mms import *

from math import pi

Lx = 2 * pi
Lz = 2 * pi

ZMAX = Lz / (2 * pi)

metric = Metric()  # Identity metric

# Define solution in terms of input x,y,z

g = sin(6 * x**2 - z)  # Constant drive for advection

f = cos(4 * x**2 + z) + sin(t) * sin(3 * x + 2 * z)

# Turn solution into real x and z coordinates
replace = [(x, metric.x / Lx), (z, metric.z / ZMAX)]

f = f.subs(replace)
g = g.subs(replace)

# Calculate time derivatives

dfdt = -bracket(g, f, metric)

# Calculate source
S = diff(f, t) - dfdt

# Differentials for boundary conditions
dfdx = diff(f, metric.x)

# Substitute back to get input x and z coordinates
replace = [(metric.x, x * Lx), (metric.z, z * ZMAX)]

g = g.subs(replace)
f = f.subs(replace)
dfdt = dfdt.subs(replace)
S = S.subs(replace)
dfdx = dfdx.subs(replace)

print("[g]")
print("solution = " + exprToStr(g))

print("\n[f]")
print("solution = " + exprToStr(f))
print("\nddx = " + exprToStr(dfdx))
print("\nsource = " + exprToStr(S))



#!/bin/python3

import os  # corelib
import glob  # corelib
import hashlib  # corelib
import base64  # corelib
import tempfile  # corelib
import subprocess  # corelib
import re  # corelib
import pathlib  # corelib
import contextlib  # corelib

try:
    import packaging.tags  # packaging
except:
    packaging = None


useLocalVersion = True
pkgname = "boutpp"
version = None


def getversion():
    """
    Get the current BOUT++ version.

    Use a version similar to setuptools_scm but fancier. Similar to cmake's
    versioning.
    """
    global version
    if version is None:
        with contextlib.suppress(KeyError):
            # 0. Check whether version is set via environment variable
            version = os.environ["BOUT_PRETEND_VERSION"]
            return version.lstrip("v")

        _bout_previous_version = "v5.2.0"
        _bout_next_version = "v5.2.1"

        try:
            try:
                # 1. Check whether we are at a tag
                version = run2("git describe --exact-match --tags HEAD").strip()
            except subprocess.CalledProcessError:
                # 2. default mode, try to derive version from previous tag
                tmp = run2(
                    f"git describe --tags --match={_bout_previous_version}"
                ).strip()
                tmp = re.sub(
                    f"{_bout_previous_version}-", f"{_bout_next_version}.dev", tmp
                )
                if useLocalVersion:
                    tmp = re.sub("-", "+", tmp)
                else:
                    tmp = re.sub("-.*", "", tmp)
                version = tmp
            with open("_version.txt", "w") as f:
                f.write(version + "\n")
        except subprocess.CalledProcessError:
            try:
                # 3. Check whether there is a _version - e.g. we have a tarball
                with open("_version.txt") as f:
                    version = f.read().strip()
            except FileNotFoundError:
                # 4. Maybe not released yet, but version already bumped?
                #    Things are messy here, so always assume useLocalVersion
                try:
                    # 4.1 us proper hash
                    hash = "g" + run2('git log -n 1 --pretty=format:"%h"')
                except subprocess.CalledProcessError:
                    # 4.2 fallback
                    hash = "unknown"
                version = _bout_previous_version + ".rc+" + hash
                with open("_version.txt", "w") as f:
                    f.write(version + "\n")
    return version.lstrip("v")


def run(cmd):
    """
    Run a command without capturing output, so it gets printed.
    """
    print(f"running `{cmd}`")
    ret = os.system(cmd)
    if ret != 0:
        raise subprocess.CalledProcessError(ret, cmd)


def run2(cmd):
    """
    Run a command and return standard-out
    """
    return subprocess.run(
        cmd, capture_output=True, shell=True, check=True
    ).stdout.decode()


def hash(fn):
    """
    Calculate the hash of a file
    """
    sha256_hash = hashlib.sha256()
    with open(fn, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return f"sha256={base64.urlsafe_b64encode(sha256_hash.digest()).decode()[:-1]}"


def size(fn):
    """
    Return the size of a file in bytes
    """
    return os.path.getsize(fn)


def gettag():
    """
    Get the platform tag
    """
    thisos = list(packaging.tags.platform_tags())[-1]
    tag = "-".join(str(next(packaging.tags.sys_tags())).split("-")[:2] + [thisos])
    return tag


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    """
    Build the wheel.

    Calls cmake internaly.
    """
    if metadata_directory:
        parse(f"{metadata_directory}/METADATA")
    opts = ""
    if config_settings is not None:
        global useLocalVersion, pkgname
        for k, v in config_settings.items():
            if k == "sdist":
                continue
            if k == "useLocalVersion":
                useLocalVersion = False
                continue
            if k == "nightly":
                useLocalVersion = False
                pkgname = "boutpp-nightly"
                continue
            if v:
                opts += f" {k}={v}"
            else:
                opts += f" {k}=ON"
    tag = gettag()
    whlname = f"{pkgname.replace('-', '_')}-{getversion()}-{tag}.whl"
    trueprefix = f"{os.getcwd()}/_wheel_install/"
    prefix = f"{trueprefix}/boutpp/"
    run(
        "cmake -S . -B _wheel_build/ -DBOUT_ENABLE_PYTHON=ON"
        + f" -DCMAKE_INSTALL_PREFIX={prefix} -DCMAKE_INSTALL_LIBDIR={prefix}"
        + f" -DCMAKE_INSTALL_PYTHON_SITEARCH={trueprefix} -DCMAKE_INSTALL_RPATH=$ORIGIN"
        + opts
    )
    run(f"cmake --build  _wheel_build/ -j {os.cpu_count()}")
    run("cmake --install _wheel_build/")
    distinfo = f"_wheel_install"
    prepare_metadata_for_build_wheel("_wheel_install", record=True)

    # Do not add --symlink as python's does not extract that as symlinks
    run(f"cd {trueprefix} ; zip  {wheel_directory}/{whlname} . -rq")
    return whlname


def build_sdist(sdist_directory, config_settings=None):
    """
    Create an archive of the code including some metadata files
    """
    print(config_settings, sdist_directory)
    enable_gz = True
    enable_xz = False
    external = {"fmt", "mpark.variant", "cpptrace"}
    if config_settings is not None:
        global useLocalVersion, pkgname
        for k, v in config_settings.items():
            if k == "sdist":
                if v == "onlygz":
                    enable_gz = True
                    enable_xz = False
                elif v == "onlyxz":
                    enable_xz = True
                    enable_gz = False
                elif v == "both":
                    enable_xz = True
                else:
                    raise ValueError(f"unknown option {v} for {k}")
            if k == "dist":
                enable_xz = True
                pkgname = "BOUT++"
                external.add("googletest")
            if k == "useLocalVersion":
                useLocalVersion = False
            if k == "nightly":
                useLocalVersion = False
                pkgname = "boutpp-nightly"
    prefix = f"{pkgname.replace('-', '_')}-{getversion()}"
    fname = f"{prefix}.tar"
    run(f"git archive HEAD --prefix {prefix}/ -o {sdist_directory}/{fname}")
    _, tmp = tempfile.mkstemp(suffix=".tar")
    for ext in sorted(external):
        run(
            f"git archive --remote=externalpackages/{ext} HEAD --prefix  {prefix}/externalpackages/{ext}/ --format=tar > {tmp}"
        )
        run(f"tar -Af {sdist_directory}/{fname} {tmp}")
        run(f"rm {tmp}")

    with open(tmp, "w") as f:
        f.write(
            f"""Metadata-Version: 2.1
Name: {pkgname}
Version: {getversion()}
"""
        )
        with open("LICENSE") as src:
            pre = "License: "
            for l in src:
                f.write(f"{pre}{l}")
                pre = "         "
        f.write("Description-Content-Type: text/markdown\n\n")
        with open("README.md") as src:
            f.write(src.read())
    run(
        f"tar --append -f {sdist_directory}/{fname} _version.txt --xform='s\\_version.txt\\{prefix}/_version.txt\\'"
    )
    run(
        f"tar --append -f {sdist_directory}/{fname} {tmp} --xform='s\\{tmp[1:]}\\{prefix}/PKG-INFO\\'"
    )

    if enable_gz:
        run(f"gzip --force --best --keep {sdist_directory}/{fname}")
        if not enable_xz:
            fname += ".gz"
    if enable_xz:
        run(f"rm {sdist_directory}/{fname}.xz -f")
        run(f"xz --best {sdist_directory}/{fname}")
        fname += ".xz"
    return fname


def get_requires_for_build_sdist(config_settings=None):
    return []


def get_requires_for_build_wheel(config_settings=None):
    return ["packaging", "cython", "jinja2", "numpy"]


def mkdir_p(path):
    return pathlib.Path(path).mkdir(parents=True, exist_ok=True)


def prepare_metadata_for_build_wheel(
    metadata_directory, config_settings=None, record=False
):
    """
    Create dist-info directory with files
    """
    if not record and os.path.isfile("PKG-INFO"):
        parse("PKG-INFO")

    thisdir = f"{pkgname.replace('-', '_')}-{getversion()}.dist-info"
    distinfo = f"{metadata_directory}/{thisdir}"
    mkdir_p(distinfo)
    with open(f"{distinfo}/METADATA", "w") as f:
        f.write(
            f"""Metadata-Version: 2.1
Name: {pkgname}
Version: {getversion()}
License-File: COPYING
"""
        )
    run(f"cp LICENSE {distinfo}/COPYING")
    run(f"cp LICENSE.GPL {distinfo}/COPYING.GPL")
    with open(f"{distinfo}/WHEEL", "w") as f:
        f.write(
            f"""Wheel-Version: 1.0
Generator: boutpp_custom_build_wheel (version {getversion()})
Root-Is-Purelib: false
Tag: {gettag()}
"""
        )

    if record:
        with open(f"{distinfo}/RECORD", "w") as f:
            for fn in glob.iglob("_wheel_install/**", recursive=True):
                if not os.path.isfile(fn):
                    continue
                fn0 = fn.removeprefix("_wheel_install/")
                if fn0 != f"{distinfo}/RECORD":
                    f.write(f"{fn0},{hash(fn)},{size(fn)}\n")
                else:
                    f.write(f"{fn0},,\n")
    return thisdir


def parse(fn):
    with open(fn) as f:
        global pkgname, version
        for line in f:
            if line.startswith("Name:"):
                pkgname = line[5:].strip()
            if line.startswith("Version:"):
                version = line[8:].strip()


def nightly():
    """
    Build the python sdist for upload to boutpp-nightly.
    """
    return build_sdist(os.getcwd() + "/dist/", dict(nightly=True))


def sdist():
    """
    Build the python sdist
    """
    return build_sdist(os.getcwd() + "/dist/")


def wheel():
    """
    Build the python binary wheel
    """
    return build_wheel(os.getcwd() + "/dist/")


def dist():
    """
    Build an archive for BOUT++ release
    """
    return build_sdist(os.getcwd(), config_settings=dict(dist=True))


def help():
    """
    Print this help
    """
    table = []
    for k, v in todos.items():
        try:
            doc = v.__doc__.strip()
            doc = " : " + doc
        except:
            doc = ""
        table.append((k, doc))
    maxkey = max([len(k) for k, _ in table])
    fmt = f"   %-{maxkey}s%s"
    print(f"{sys.argv[0]} [command] [command]")
    for row in table:
        print(fmt % row)


def printVersion():
    """
    print the version
    """
    print(getversion())


todos = dict(
    nightly=nightly,
    sdist=sdist,
    wheel=wheel,
    dist=dist,
    version=printVersion,
    help=help,
)
todos.update(
    {
        "--help": help,
        "-?": help,
        "-h": help,
    }
)

if __name__ == "__main__":
    import sys

    for todo in sys.argv[1:]:
        if todo not in todos:
            help()
            sys.exit(1)
        todos[todo]()



import csv


def read_file(filename):
    reader = csv.reader(open(filename, "r"), delimiter="\t", skipinitialspace=True)

    # Skip header
    for _, _ in zip(range(4), reader):
        continue

    from collections import OrderedDict

    case_lines = OrderedDict()  # {}
    for line in reader:
        if line == []:
            break
        case_lines[line[0].rstrip(".")] = line[1]

    titles = next(reader)
    cases_weak = {col.strip(): [] for col in titles[:-1]}

    for line in reader:
        if line == []:
            break
        for title, col in zip(titles, line[:-1]):
            cases_weak[title].append(float(col))

    axis = cases_weak["Local grid"]
    data = [cases_weak[x] for x in case_lines]
    labels = [case_lines[x] for x in case_lines]
    return {"axis": axis, "data": data, "labels": labels}


def getScan(baseName="timing_{n}.txt", nthreads=(1, 2, 4, 8, 16, 32)):
    from numpy import zeros

    dataS = []
    for n in nthreads:
        f = baseName.format(n=n)
        dataS.append(read_file(f))

    nnt = len(dataS)
    nlines = len(dataS[0]["data"])
    nval = len(dataS[0]["data"][0])
    rawDat = zeros([nnt, nval, nlines])

    for i, dat in enumerate(dataS):
        print(len(dat["data"]))

        rawDat[i, :, :] = dat["data"]

    axes = [nthreads, dataS[0]["axis"]]

    return axes, rawDat, dataS[0]["labels"]



#!/bin/python3
# -*- coding: utf-8

#!##################################################################
#
#       2D blob simulations
#
#  Copyright:   NR Walkden, B Dudson, D Bold;  2012, 2017, 2018
#
###################################################################

import boutpp as bc
from numpy import sqrt
from boutpp import bracket, DDZ, Delp2, PhysicsModel
import sys
import os


class Blob2D(PhysicsModel):
    def init(self, restart):
        self.mesh = bc.Mesh.getGlobal()
        self.n = bc.Field3D.fromMesh(self.mesh)
        self.omega = bc.Field3D.fromMesh(self.mesh)

        self.phiSolver = bc.Laplacian()

        options = bc.Options.root("model")
        # Temperature in eV
        Te0 = options.get("Te0", 30)
        e = options.get("e", 1.602e-19)
        m_i = options.get("m_i", 2 * 1.667e-27)
        m_e = options.get("m_e", 9.11e-31)

        # Viscous diffusion coefficient
        self.D_vort = options.get("D_vort", 0)
        # Density diffusion coefficient
        self.D_n = options.get("D_n", 0)

        # Radius of curvature [m]
        self.R_c = options.get("R_c", 1.5)
        # Parallel connection length [m]
        self.L_par = options.get("L_par", 10)

        # Value of magnetic field strength [T]
        B0 = options.get("B0", 0.35)

        # System option switches

        # Include compressible ExB term in density equation
        self.compressible = options.get("compressible", False)
        # Use Boussinesq approximation in vorticity
        self.boussinesq = options.get("boussinesq", True)
        # Sheath closure
        self.sheath = options.get("sheath", True)

        Omega_i = e * B0 / m_i  # Cyclotron Frequency
        c_s = sqrt(e * Te0 / m_i)  # Bohm sound speed
        self.rho_s = c_s / Omega_i  # Bohm gyro-radius

        print(
            "\n\n\t----------Parameters: ------------ \n\tOmega_i = %e /s,\n\t"
            "c_s = %e m/s,\n\trho_s = %e m\n" % (Omega_i, c_s, self.rho_s)
        )

        # Calculate delta_*, blob size scaling
        print(
            "\tdelta_* = rho_s * (dn/n) * %e "
            % (pow(self.L_par * self.L_par / (self.R_c * self.rho_s), 1.0 / 5))
        )

        # /************ Create a solver for potential ********/

        opts_boussinesq = bc.Options.root("phiBoussinesq")
        opts_non_boussinesq = bc.Options.root("phiSolver")

        if self.boussinesq:
            # BOUT.inp section "phiBoussinesq"
            opts_used = opts_boussinesq
            opts_unused = opts_non_boussinesq
        else:
            # BOUT.inp section "phiSolver"
            opts_used = opts_non_boussinesq
            opts_unused = opts_boussinesq

        self.phiSolver = bc.Laplacian(opts_used)
        opts_unused.setConditionallyUsed()

        # Starting guess for first solve (if iterative)
        self.phi = bc.create3D("0")

        # /************ Tell BOUT++ what to solve ************/

        self.solve_for(n=self.n, omega=self.omega)
        # model.save_repeat(phi=phi)
        # model.save_once(rho_s=rho_s,c_s=c_s,Omega_i=Omega_i)

    def rhs(self, time):
        # Run communications
        ######################################
        self.mesh.communicate(self.n, self.omega)

        # Invert div(n grad(phi)) = grad(n) grad(phi) + n Delp_perp^2(phi) = omega
        ######################################
        # Set the time derivative by adding/... to it
        # make sure to never overwrite it
        # ddt_n = bla does NOT set the time derivative
        ddt_n = self.n.ddt()
        ddt_n.set(0)
        if not self.boussinesq:
            # Including full density in vorticit inversion
            # Update the 'C' coefficient. See invert_laplace.hxx
            self.phiSolver.setCoefC(n)
            # Use previous solution as guess
            self.phi = self.phiSolver.solve(omega / n, self.phi)
        else:
            # Background density only (1 in normalised units)
            self.phi = self.phiSolver.solve(self.omega, self.phi)

        self.mesh.communicate(self.phi)

        # Density Evolution
        # /

        # ExB term
        ddt_n += -bracket(self.phi, self.n, "BRACKET_SIMPLE")
        # Curvature term
        ddt_n += 2 * DDZ(self.n) * (self.rho_s / self.R_c)
        # Diffusion term
        ddt_n += self.D_n * Delp2(self.n)

        if self.compressible:
            # ExB Compression term
            ddt_n -= 2 * self.n * DDZ(self.phi) * (self.rho_s / self.R_c)

        if self.sheath:
            # Sheath closure
            ddt_n += (
                self.n * self.phi * (self.rho_s / self.L_par)
            )  # - (n - 1)*(rho_s/L_par)

        # Vorticity evolution
        # /
        # ExB term
        ddt_omega = -bracket(self.phi, self.omega, "BRACKET_SIMPLE")
        ddt_omega += 2 * DDZ(self.n) * (self.rho_s / self.R_c) / self.n
        # Viscous diffusion term
        ddt_omega += self.D_vort * Delp2(self.omega) / self.n

        if self.sheath:
            ddt_omega += self.phi * (self.rho_s / self.L_par)
        # other option to set time derivaitve:
        # create a field and set it in the end
        self.omega.ddt(ddt_omega)


# Ensure the blob folder exists
def ensure_blob():
    if not os.path.isdir("blob"):
        print("Setting up folder blob for simulation ...")
        os.mkdir("blob")
    if not os.path.exists("blob/BOUT.inp"):
        with open("blob/BOUT.inp", "w") as f:
            f.write(
                """\
# settings file for BOUT++
#
# Blob simulation in a 2D slab
#
# This case has blob size
#
# delta = 0.3*256 ~ 10 * delta_*


# settings used by the core code

nout = 50      # number of time-steps
timestep = 50  # time between outputs [1/wci]


MXG = 2      # Number of X guard cells
MYG = 0      # No y derivatives, so no guard cells needed in y

[mesh]

nx = 260    # Note: 4 guard cells
ny = 1
nz = 256

dx = 0.3      # Grid spacing [rho_s]
dz = 0.3

##################################################
# derivative methods

[mesh:ddx]

first = C2
second = C2
upwind = W3

[mesh:ddy]

first = C2
second = C2
upwind = W3

[mesh:ddz]

first = C2
second = C2
upwind = W3

###################################################
# Time-integration solver

[solver]

atol = 1e-10  # absolute tolerance
rtol = 1e-05   # relative tolerance
mxstep = 10000  # Maximum internal steps per output

###################################################
# Electrostatic potential solver
# These options are used if boussinesq = false

[phiSolver]
type = petsc  # Needed if Boussinesq = false
pctype = user  # Preconditioning type

fourth_order = true  # 4th order or 2nd order

# 0  = Zero value
# 10 = Zero gradient AC inner & outer
# 15 = Zero gradient AC and DC
# 768 = Zero laplace inner & outer

[phiSolver:precon]  # Preconditioner (if pctype=user)
filter = 0.0     # Must not filter solution
flags = 49152  # set_rhs i.e. identity matrix in boundaries

###################################################
# Electrostatic potential solver (Boussinesq)

[phiBoussinesq]
# By default type is tri (serial) or spt (parallel)

##################################################
# general settings for the model

[model]

Te0 = 5    # Electron Temperature (eV)

compressible = false  # Compressibility?

boussinesq = true  # Boussinesq approximation (no perturbed n in vorticity)

D_vort = 1e-06  # Viscosity
D_n = 1e-06    # Diffusion

R_c = 1.5  # Radius of curvature (m)

# settings for individual variables
# The section "All" defines default settings for all variables
# These can be overridden for individual variables in
# a section of that name.

[all]
scale = 0.0 # default size of initial perturbations

bndry_all = neumann # Zero-gradient on all boundaries

[n]  # Density
scale = 1.0 # size of perturbation

height = 0.5
width = 0.05

function = 1 + height * exp(-((x-0.25)/width)^2 - ((z/(2*pi) - 0.5)/width)^2)
"""
            )


if __name__ == "__main__":
    ensure_blob()

    bc.init("-d blob".split(" ") + sys.argv[1:])

    # Create an instance
    blob2d = Blob2D()
    # Start the simulation
    blob2d.solve()



#
# SNB model example
# A small sinusoidal temperature perturbation is added to a periodic 1D domain

import numpy as np
import matplotlib.pyplot as plt

from boututils.run_wrapper import launch_safe, build_and_log
from boutdata.collect import collect

build_and_log("Sinusoidal SNB")

# Electron temperature in eV
Telist = 10 ** np.linspace(0, 3, 20)

# Electron density in m^-3
Ne = 1e20

# Length of the domain in m
length = 1.0

c = 299792458
mu0 = 4.0e-7 * np.pi
e0 = 1 / (c * c * mu0)
qe = 1.602176634e-19
me = 9.10938356e-31

thermal_speed = np.sqrt(2.0 * qe * Telist / me)
Y = (qe**2 / (e0 * me)) ** 2 / (4 * np.pi)
coulomb_log = 6.6 - 0.5 * np.log(Ne * 1e-20) + 1.5 * np.log(Telist)
lambda_ee_T = thermal_speed**4 / (Y * Ne * coulomb_log)

beta_max_list = [5, 10, 20, 40]
colors = ["k", "b", "g", "r"]

ngroups_list = [20, 40, 80]
syms = ["x", "o", "D"]

for beta_max, color in zip(beta_max_list, colors):
    for ngroups, sym in zip(ngroups_list, syms):
        flux_ratio = []
        for Te in Telist:
            cmd = './conduction-snb "Te={0}+0.01*sin(y)" Ne={1} mesh:length={2} snb:beta_max={3} snb:ngroups={4}'.format(
                Te, Ne, length, beta_max, ngroups
            )

            # Run the case
            s, out = launch_safe(cmd, nproc=1, mthread=1, pipe=True)

            div_q = collect("Div_Q", path="data").ravel()
            div_q_SH = collect("Div_Q_SH", path="data").ravel()

            # Get the index of maximum S-H heat flux
            ind = np.argmax(div_q_SH)

            flux_ratio.append(div_q[ind] / div_q_SH[ind])

        plt.plot(
            lambda_ee_T / length,
            flux_ratio,
            "-" + sym + color,
            label=r"$\beta_{{max}}={0}, N_g={1}$".format(beta_max, ngroups),
        )

plt.legend()
plt.xlabel(r"$\lambda_{ee,T} / L$")
plt.ylabel(r"$q / q_{SH}$")
plt.xscale("log")

plt.savefig("snb-sinusoidal.png")
plt.show()



#! /usr/bin/env python

# MPark.Variant
#
# Copyright Michael Park, 2017
#
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE.md or copy at http://boost.org/LICENSE_1_0.txt)

import os.path
import re
import subprocess
import sys

# Prints a single header version of `include/mpark/variant.hpp` to stdout.

processed = []

def process(header):
  result = ''
  with open(header, 'r') as f:
    for line in f:
      p = re.compile('^#include "(.+)"')
      m = p.match(line)
      if m is None:
        result += line
      else:
        g = m.group(1)
        include = os.path.normpath(os.path.join(os.path.dirname(header), g))
        if include not in processed:
          result += process(include)
          result += '\n'
          processed.append(include)
  return result

root = subprocess.check_output(['git', 'rev-parse', '--show-toplevel']).strip()
result = process(os.path.join(root, 'include/mpark/variant.hpp'))

sys.stdout.write(result)



#!/bin/python3
import boutpp as bc

bc.init("mesh:n=48")


class Model(bc.PhysicsModel):
    def init(self, restart):
        self.dens = bc.create3D("sin(x)")
        self.solve_for(n=self.dens)

    def rhs(self, time):
        self.dens.ddt(bc.DDX(self.dens))


model = Model()
model.solve()



"""
@brief Convert an equilibrium set of flux surfaces into a BOUT++ grid


"""

from __future__ import print_function

from numpy import max

from boututils.datafile import DataFile


def grid2bout(input, output="bout.grd.nc"):
    # List variables needed and dimensions
    dimensions = {
        "nx": 0,
        "ny": 0,
        "psi": 1,
        "mu0p": 1,
        "mu0pprime": 1,
        "f": 1,
        "ffprime": 1,
        "qsafe": 1,
        "Rxy": 2,
        "Zxy": 2,
    }

    # Open the input file
    infile = Datafile(input)

    # Create an empty dictionary for the variables
    vars = {}

    # Check the required variables
    for k, v in list(dimensions.items()):
        nd = infile.ndims(k)
        if nd == None:
            print("ERROR: Variable missing: " + k)
            return False
        if v != nd:
            print("ERROR: Variable '" + k + "' has wrong number of dimensions")
            return False

        # Read the variables
        vars[k] = f.read(k)

    # Read the size of the grid
    nx = vars["nx"]
    ny = vars["ny"]

    # Location of grid points
    Rxy = vars["Rxy"]
    Zxy = vars["Zxy"]

    # Check if the last point is the same as the first
    dl = (Rxy[:, ny - 1] - Rxy[:, 0]) ** 2 + (Zxy[:, ny - 1] - Zxy[:, 0]) ** 2
    if dl.max() < 1.0e-4:
        print("**Last poloidal point duplicates the first. Removing...")



#!/usr/bin/env python

#
# Generate an input mesh
#

from boututils.datafile import DataFile  # Wrapper around NetCDF4 libraries

nx = 5  # Minimum is 5: 2 boundary, one evolved
ny = 64  # Minimum 5. Should be divisible by number of processors (so powers of 2 nice)

f = DataFile()
f.open("conduct_grid.nc", create=True)

f.write("nx", nx)
f.write("ny", ny)

f.close()



from __future__ import division
from __future__ import print_function

from boutdata.mms import Metric, sin, cos, Div_par, Grad_par, exprToStr, diff, y, t
from math import pi

# Length of the y domain
Ly = 10.0

# metric tensor
metric = Metric()  # Identity

# Define solution in terms of input x,y,z

n = 1 + 0.1 * sin(2 * y - t)
p = 1 + 0.1 * cos(3 * y + t)
nv = 0.1 * sin(y + 2 * t)

# Turn solution into real x and z coordinates
replace = [(y, metric.y * 2 * pi / Ly)]

n = n.subs(replace)
p = p.subs(replace)
nv = nv.subs(replace)

##############################
# Calculate time derivatives

v = nv / n
gamma = 5.0 / 3

# Density equation
dndt = -Div_par(nv)

# Pressure equation
dpdt = -Div_par(p * v) - (gamma - 1.0) * p * Div_par(v)

# Momentum equation
dnvdt = -Div_par(nv * v) - Grad_par(p)

#############################
# Calculate sources

Sn = diff(n, t) - dndt
Sp = diff(p, t) - dpdt
Snv = diff(nv, t) - dnvdt

# Substitute back to get input y coordinates
replace = [(metric.y, y * Ly / (2 * pi))]

n = n.subs(replace)
p = p.subs(replace)
nv = nv.subs(replace)

Sn = Sn.subs(replace)
Sp = Sp.subs(replace)
Snv = Snv.subs(replace)

print("[n]")
print("solution = " + exprToStr(n))
print("\nsource = " + exprToStr(Sn))

print("\n[p]")
print("solution = " + exprToStr(p))
print("\nsource = " + exprToStr(Sp))

print("\n[nv]")
print("solution = " + exprToStr(nv))
print("\nsource = " + exprToStr(Snv))



#!/usr/bin/env python3
"""Code-generator for arithmetic operators on Field2Ds/Field3Ds

This uses the jinja template in gen_fieldops.jinja to generate code
for the arithmetic operators, and prints to stdout.

The `Field` class provides some helper functions for determining how to
pass a variable by reference or pointer, and how to name arguments in
function signatures. This allows us to push some logic into the
templates themselves.

"""

from __future__ import print_function

from builtins import object

import argparse
from collections import OrderedDict
import contextlib
from copy import deepcopy as copy
import itertools
import sys

try:
    import jinja2
except ImportError:
    raise ImportError(
        'Missing Python module "jinja2". See "Field2D/Field3D Arithmetic '
        'Operators" in the BOUT++ user manual for more information'
    )


# This allows us to open either a file or stdout with the same code
@contextlib.contextmanager
def smart_open(filename, mode="r"):
    """Open stdin or stdout using a contextmanager

    From: http://stackoverflow.com/a/29824059/2043465
    """
    if filename == "-":
        if mode is None or mode == "" or "r" in mode:
            fh = sys.stdin
        else:
            fh = sys.stdout
    else:
        fh = open(filename, mode)
    try:
        yield fh
    finally:
        if filename != "-":
            fh.close()


# The arthimetic operators
# OrderedDict to (try to) ensure consistency between python 2 & 3
operators = OrderedDict(
    [
        ("*", "multiplication"),
        ("/", "division"),
        ("+", "addition"),
        ("-", "subtraction"),
    ]
)

header = """// This file is autogenerated - see gen_fieldops.py
#include <bout/field2d.hxx>
#include <bout/field3d.hxx>
#include <bout/globals.hxx>
#include <bout/interpolation.hxx>
#include <bout/mesh.hxx>
#include <bout/region.hxx>
"""


class Field(object):
    """Abstracts over BoutReals and Field2D/3D/Perps

    Provides some helper functions for writing function signatures and
    passing data

    """

    def __init__(
        self,
        field_type,
        dimensions,
        name=None,
        index_var=None,
        jz_var="jz",
        mixed_base_ind_var="base_ind",
    ):
        # C++ type of the field, e.g. Field3D
        self.field_type = field_type
        # array: dimensions of the field
        self.dimensions = dimensions
        # name of this field
        self.name = name
        # Name of the indexing variable
        self.index_var = index_var
        # Name of jz variable
        self.jz_var = jz_var
        self.mixed_base_ind_var = mixed_base_ind_var
        # Note region_type isn't actually used currently but
        # may be useful in future.
        if self.field_type == "Field3D":
            self.region_type = "3D"
        elif self.field_type == "Field2D":
            self.region_type = "2D"
        else:
            self.region_type = "_INVALID_"

    @property
    def passByReference(self):
        """Returns "Type& name", except if field_type is BoutReal,
        in which case just returns "Type name"

        """
        return "{self.field_type}{ref} {self.name}".format(
            self=self, ref="&" if self.field_type != "BoutReal" else ""
        )

    @property
    def index(self):
        """Returns "[{index_var}]", except if field_type is BoutReal,
        in which case just returns ""

        """
        if self.field_type == "BoutReal":
            return "{self.name}".format(self=self)
        else:
            return "{self.name}[{self.index_var}]".format(self=self)

    @property
    def mixed_index(self):
        """Returns "[{index_var} + {jz_var}]" if field_type is Field3D,
        self.index if Field2D or just returns "" for BoutReal

        """
        if self.field_type == "BoutReal":
            return "{self.name}".format(self=self)
        elif self.field_type == "Field3D":
            return "{self.name}[{self.mixed_base_ind_var} + {self.jz_var}]".format(
                self=self
            )
        else:  # Field2D
            return "{self.name}[{self.index_var}]".format(self=self)

    @property
    def base_index(self):
        """Returns "[{mixed_base_ind_var}]" if field_type is Field3D, Field2D or FieldPerp
        or just returns "" for BoutReal

        """
        if self.field_type == "BoutReal":
            return "{self.name}".format(self=self)
        else:
            return "{self.name}[{self.mixed_base_ind_var}]".format(self=self)

    def __eq__(self, other):
        try:
            return self.field_type == other.field_type
        except AttributeError:
            return self.field_type == other

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "Field({}, {}, {})".format(self.field_type, self.dimensions, self.name)

    def __str__(self):
        return self.field_type


def returnType(f1, f2):
    """Determine a suitable return type, by seeing which field is 'larger'."""
    if f1 == f2:
        return copy(f1)
    elif f1 == "BoutReal":
        return copy(f2)
    elif f2 == "BoutReal":
        return copy(f1)
    elif f1 == "FieldPerp" or f2 == "FieldPerp":
        return copy(fieldPerp)
    else:
        return copy(field3D)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate code for the Field arithmetic operators"
    )
    # By default write to stdout
    parser.add_argument(
        "--filename", default="-", help="Write output to FILENAME instead of stdout"
    )
    # By default use OpenMP enabled loops but allow to disable
    parser.add_argument(
        "--no-openmp",
        action="store_false",
        default=False,
        dest="noOpenMP",
        help="Don't use OpenMP compatible loops",
    )

    args = parser.parse_args()

    # Setup
    index_var = "index"
    jz_var = "jz"
    mixed_base_ind_var = "base_ind"
    region_name = '"RGN_ALL"'

    if args.noOpenMP:
        region_loop = "BOUT_FOR_SERIAL"
    else:
        region_loop = "BOUT_FOR"

    # Declare what fields we currently support:
    # Field perp is currently missing
    field3D = Field(
        "Field3D",
        ["x", "y", "z"],
        index_var=index_var,
        jz_var=jz_var,
        mixed_base_ind_var=mixed_base_ind_var,
    )
    field2D = Field(
        "Field2D",
        ["x", "y"],
        index_var=index_var,
        jz_var=jz_var,
        mixed_base_ind_var=mixed_base_ind_var,
    )
    fieldPerp = Field(
        "FieldPerp",
        ["x", "z"],
        index_var=index_var,
        jz_var=jz_var,
        mixed_base_ind_var=mixed_base_ind_var,
    )
    boutreal = Field(
        "BoutReal",
        [],
        index_var=index_var,
        jz_var=jz_var,
        mixed_base_ind_var=mixed_base_ind_var,
    )

    fields = [field3D, field2D, fieldPerp, boutreal]

    with smart_open(args.filename, "w") as f:
        f.write(header)
        f.write("\n")

    env = jinja2.Environment(loader=jinja2.FileSystemLoader("."), trim_blocks=True)

    template = env.get_template("gen_fieldops.jinja")

    for lhs, rhs in itertools.product(fields, fields):
        # We don't have to define BoutReal BoutReal operations
        if lhs == rhs == "BoutReal":
            continue
        rhs = copy(rhs)
        lhs = copy(lhs)

        # The output of the operation. The `larger` of the two fields.
        out = returnType(rhs, lhs)
        out.name = "result"
        lhs.name = "lhs"
        rhs.name = "rhs"

        for operator, operator_name in operators.items():
            template_args = {
                "operator": operator,
                "operator_name": operator_name,
                #
                "out": out,
                "lhs": lhs,
                "rhs": rhs,
                #
                "region_loop": region_loop,
                "region_name": region_name,
                #
                "index_var": index_var,
                "mixed_base_ind": mixed_base_ind_var,
                "jz_var": jz_var,
            }

            with smart_open(args.filename, "a") as f:
                f.write(template.render(**template_args))
                f.write("\n")



from __future__ import print_function
from __future__ import division
from past.utils import old_div

#
# Generate the test case using SymPy
#
# Equations are:
#
#
# Delp2(phi) = vort + Sphi

from boutdata.mms import *

from sympy import log, Wild

####


def C(f):
    """Curvature"""
    return bxcvz * diff(f, metric.z)


#######################################
# Inputs, essentially as in BOUT.inp

# Switches
ionvis = False  # Ion Viscosity
Ti = 10  # Ion temperature for viscosity calculation
elecvis = False  # Electron viscosity
resistivity = True

parallel = True  # Parallel dynamics

# Parameters
Tnorm = 3  # Electron Temperature (eV)
Nnorm = 1e19  # Background plasma density (m^-3)
Bnorm = 0.1  # Magnetic field [T]
AA = 0.1  # Ion atomic mass

#######################################
#

mi_me = AA * 1.67262158e-27 / 9.109e-31
beta_e = qe * Tnorm * Nnorm / (old_div(Bnorm**2, mu0))

# Normalisation parameters
Cs0 = sqrt(qe * Tnorm / (AA * Mp))
Omega_ci = qe * Bnorm / (AA * Mp)
rho_s0 = old_div(Cs0, Omega_ci)
Coulomb = 6.6 - 0.5 * log(Nnorm * 1e-20) + 1.5 * log(Tnorm)
tau_e0 = old_div(
    1.0, (2.91e-6 * (old_div(Nnorm, 1e6)) * Coulomb * Tnorm ** (old_div(-3.0, 2)))
)
# Input settings

ZMAX = 1e-3

MXG = 2
MYG = 2

bxcvz = 100  # Curvature

Rxy = 1.5  # Major radius
Bpxy = 0.35  # Poloidal magnetic field
Bxy = 0.35  # Total magnetic field
Btxy = 0.0  # Toroidal magnetic field
hthe = 0.1  # Poloidal arc length

sinty = 0.0
sbp = 1.0

dx = 2e-5
dy = 1e-3

nx = 132
ny = 64

estatic = True

# Normalise

Rxy /= rho_s0
Bpxy /= Bnorm
Btxy /= Bnorm
Bxy /= Bnorm
hthe /= rho_s0

dx /= rho_s0**2 * Bnorm

bxcvz *= rho_s0**2

# Define a metric
metric = Metric()  # Identity

Lx = dx * (nx - 2.0 * MXG)  # Size of the X domain
Ly = dy * ny  # Size of the Y domain

metric.g11 = (Rxy * Bpxy) ** 2
metric.g22 = old_div(1.0, (hthe**2))
metric.g33 = (sinty**2) * metric.g11 + old_div((Bxy**2), metric.g11)
metric.g12 = 0.0
metric.g13 = -sinty * metric.g11
metric.g23 = -sbp * Btxy / (hthe * Bpxy * Rxy)

metric.J = old_div(hthe, Bpxy)
B = metric.B = Bxy

metric.g_11 = old_div(1.0, metric.g11) + ((sinty * Rxy) ** 2)
metric.g_22 = (Bxy * hthe / Bpxy) ** 2
metric.g_33 = Rxy * Rxy
metric.g_12 = sbp * Btxy * hthe * sinty * Rxy / Bpxy
metric.g_13 = sinty * Rxy * Rxy
metric.g_23 = sbp * Btxy * hthe * Rxy / Bpxy


print("dx = %e, dy = %e" % (dx, dy))
print("g11 = %e, g22 = %e, g33 = %e" % (metric.g11, metric.g22, metric.g33))
print("g12 = %e, g23 = %e" % (metric.g12, metric.g23))
print("g_11 = %e, g_22 = %e, g_33 = %e" % (metric.g_11, metric.g_22, metric.g_33))
print("g_12 = %e, g_23 = %e" % (metric.g_12, metric.g_23))

# Define the manufactured solution in terms of input x,y and z

Ne = 0.9 + 0.9 * x + 0.5 * cos(t) * sin(5.0 * x**2 - z) + 0.01 * sin(y - z)
Te = 1 + 0.5 * cos(t) * cos(3 * x**2 - 2 * z) + 0.005 * sin(y - z) * sin(t)
Vort = 2.0 * sin(2 * t) * cos(x - z + 4 * y)
VePsi = cos(1.5 * t) * (2.0 * sin((x - 0.5) ** 2 + z) + 0.05 * cos(3 * x**2 + y - z))
Vi = -0.01 * cos(7 * t) * cos(3 * x**2 + 2 * y - 2 * z)

# phi = sin(z - x + t)*sin(2.*pi*x)
phi = (sin(z - x + t) + 0.001 * cos(y - z)) * sin(
    2.0 * pi * x
)  # Must satisfy Dirichlet BCs for now
psi = sin(pi * x) * (
    0.5 * x - 0.1 * cos(7 * t) * sin(3.0 * x**2 + y - z)
)  # Must satisfy Dirichlet BCs for now

# Substitute to get in terms of actual x,y,z coordinates

replace = [
    (x, old_div(metric.x, Lx)),
    (y, 2 * pi * metric.y / Ly),
    (z, old_div(metric.z, ZMAX)),
]

Ne = Ne.subs(replace)
Te = Te.subs(replace)
Vort = Vort.subs(replace)
VePsi = VePsi.subs(replace)
Vi = Vi.subs(replace)
phi = phi.subs(replace)
psi = psi.subs(replace)

# Calculate RHS function

Pe = Ne * Te
if estatic:
    Ve = VePsi
else:
    Ve = VePsi - 0.5 * mi_me * beta_e * psi

print("mi_me = ", mi_me)
print("beta_e = ", beta_e)
print("mi_me*beta_e = ", mi_me * beta_e)

Gi = 0.0
Ge = 0.0

tau_e = Omega_ci * tau_e0 * (Te**1.5) / Ne
# Normalised collision time

nu = old_div(1.0, (1.96 * Ne * tau_e * mi_me))

### Electron density

dNedt = -bracket(phi, Ne, metric) + (old_div(2, B)) * (C(Pe) - Ne * C(phi))

if parallel:
    dNedt -= Ne * Grad_par(Ve, metric) + Vpar_Grad_par(Ve, Ne, metric)

### Electron temperature

dTedt = -bracket(phi, Te, metric) + (old_div(4.0, 3)) * (old_div(Te, B)) * (
    (old_div(7.0, 2)) * C(Te) + (old_div(Te, Ne)) * C(Ne) - C(phi)
)

if parallel:
    dTedt -= Vpar_Grad_par(Ve, Te, metric)
    dTedt += (
        (old_div(2.0, 3.0))
        * Te
        * (
            0.71 * Grad_par(Vi, metric)
            - 1.71 * Grad_par(Ve, metric)
            + 0.71 * (Vi - Ve) * Grad_par(Ne, metric) / Ne
        )
    )

### Vorticity

dVortdt = -bracket(phi, Vort, metric) + 2.0 * B * C(Pe) / Ne + B * C(Gi) / (3.0 * Ne)

if parallel:
    dVortdt += -Vpar_Grad_par(Vi, Vort, metric) + B**2 * (
        Grad_par(Vi - Ve, metric) + (Vi - Ve) * Grad_par(Ne, metric) / Ne
    )

### Parallel Ohm's law

if parallel:
    dVePsidt = (
        -bracket(phi, Ve, metric)
        - Vpar_Grad_par(Ve, Ve, metric)
        - mi_me * (old_div(2.0, 3.0)) * Grad_par(Ge, metric)
        - mi_me * nu * (Ve - Vi)
        + mi_me * Grad_par(phi, metric)
        - mi_me * (Te * Grad_par(Ne, metric) / Ne + 1.71 * Grad_par(Te, metric))
    )
else:
    dVePsidt = 0.0

### Parallel ion velocity

if parallel:
    dVidt = (
        -bracket(phi, Vi, metric)
        - Vpar_Grad_par(Vi, Vi, metric)
        - (old_div(2.0, 3.0)) * Grad_par(Gi, metric)
        - (Grad_par(Te, metric) + Te * Grad_par(Ne, metric) / Ne)
    )
else:
    dVidt = 0.0

#######################

# Create list of all evolving variables
vars = [
    (Ne, dNedt, "Ne"),
    (Te, dTedt, "Te"),
    (Vort, dVortdt, "Vort"),
    (VePsi, dVePsidt, "VePsi"),
    (Vi, dVidt, "Vi"),
]

replace = [(metric.x, x * Lx), (metric.y, y * Ly / (2 * pi)), (metric.z, z * ZMAX)]

# Loop over variables and print solution, source etc.
for v, dvdt, name in vars:
    # Calculate source
    S = diff(v, t) - dvdt

    dvdx = diff(v, metric.x)
    dvdy = diff(v, metric.y)

    # Substitute back to get in terms of x,y,z
    v = v.subs(replace)
    dvdx = dvdx.subs(replace)
    dvdy = dvdy.subs(replace)
    S = S.subs(replace)

    print("[" + name + "]")
    print("solution = " + exprToStr(trySimplify(v)))
    print("\nddx     = " + exprToStr(trySimplify(dvdx)))
    print("\nddy     = " + exprToStr(trySimplify(dvdy)))
    print("\nsource   = " + exprToStr(S))

# Potential
Sphi = Delp2(phi, metric) - Vort

print("\n[phi]")
print("\nsolution = " + exprToStr(phi.subs(replace)))
print("\nsource   = " + exprToStr(Sphi.subs(replace)))


# print "\n\nDelp2 phi = ", Delp2(phi, metric).subs(replace)

if not estatic:
    Spsi = Delp2(psi, metric) - 0.5 * Ne * mi_me * beta_e * psi - Ne * (Vi - VePsi)
    print("\n[psi]")
    print("\nsolution = " + exprToStr(psi.subs(replace)))
    print("\nsource = " + exprToStr(Spsi.subs(replace)))


##########################################
# Check magnitudes of individual terms


def exprmag(expr):
    """
    Estimate the magnitude of an expression

    """
    # First put into the BOUT.inp x,y,z
    expr = expr.subs(replace)

    # Replace all sin, cos with 1
    any = Wild("a")  # Wildcard
    expr = expr.replace(sin(any), 1.0)
    expr = expr.replace(cos(any), 1.0)

    # Pick maximum values of x,y,z
    expr = expr.subs(x, 1.0)
    expr = expr.subs(y, 2.0 * pi)
    expr = expr.subs(z, 2.0 * pi)

    return expr.evalf()


print("\n\n########################################")
print("\n\nDensity terms:")
print("  bracket : ", exprmag(-bracket(phi, Ne, metric)), "\n")
print("  (2/B)*C(Pe) : ", exprmag((old_div(2, B)) * C(Pe)), "\n")
print("  (2/B)*Ne*C(phi) : ", exprmag((old_div(2, B)) * Ne * C(phi)), "\n")
print("  Ne*Grad_par(Ve) :", exprmag(Ne * Grad_par(Ve, metric)), "\n")
print("  Vpar_Grad_par(Ve, Ne)", exprmag(Vpar_Grad_par(Ve, Ne, metric)))

print("\n\n########################################")
print("\n\nTemperature terms:")
print("  bracket : ", exprmag(-bracket(phi, Te, metric)), "\n")
print(
    "  C(Te)   : ",
    exprmag((old_div(4.0, 3)) * (old_div(Te, B)) * (old_div(7.0, 2)) * C(Te)),
    "\n",
)
print(
    "  (Te/Ne)*C(Ne) : ",
    exprmag((old_div(4.0, 3)) * (old_div(Te, B)) * (old_div(Te, Ne)) * C(Ne)),
    "\n",
)
print("  C(phi)  : ", exprmag(-(old_div(4.0, 3)) * (old_div(Te, B)) * C(phi)), "\n")
print("  Vpar_Grad_par(Ve, Te)", exprmag(Vpar_Grad_par(Ve, Te, metric)), "\n")
print(
    "  (2./3.)*Te*( 0.71*Grad_par(Vi)",
    exprmag((old_div(2.0, 3.0)) * Te * 0.71 * Grad_par(Vi, metric)),
    "\n",
)
print(
    "  (2./3.)*Te*1.71*Grad_par(Ve)",
    exprmag(-(old_div(2.0, 3.0)) * Te * 1.71 * Grad_par(Ve, metric)),
    "\n",
)
print(
    "  (2./3.)*Te*0.71*(Vi-Ve)*Grad_par(log(Ne))",
    exprmag((old_div(2.0, 3.0)) * Te * 0.71 * (Vi - Ve) * Grad_par(log(Ne), metric)),
    "\n",
)


print("\n\n########################################")
print("\n\nVorticity terms:")
print("  bracket : ", exprmag(-bracket(phi, Vort, metric)), "\n")
print("  C(Pe)   : ", exprmag(2.0 * B * C(Pe) / Ne), "\n")
print("  C(Gi)   : ", exprmag(B * C(Gi) / (3.0 * Ne)), "\n")
print("  Vpar_Grad_par(Vi, Vort) : ", exprmag(-Vpar_Grad_par(Vi, Vort, metric)), "\n")
print("  B**2*Grad_par(Vi - Ve) :", exprmag(B**2 * Grad_par(Vi - Ve, metric)), "\n")
print(
    "  B**2*(Vi - Ve)*Grad_par(log(Ne)) :",
    exprmag(B**2 * (Vi - Ve) * Grad_par(log(Ne), metric)),
    "\n",
)

print("\n\n########################################")
print("\n\nOhm's law terms:")

print("  bracket : ", exprmag(-bracket(phi, Ve, metric)), "\n")
print("  Vpar_Grad_par(Ve, Ve) : ", exprmag(-Vpar_Grad_par(Ve, Ve, metric)), "\n")
print(
    "  mi_me*(2./3.)*Grad_par(Ge)",
    exprmag(-mi_me * (old_div(2.0, 3.0)) * Grad_par(Ge, metric)),
    "\n",
)
print("  mi_me*nu*(Ve - Vi)", exprmag(-mi_me * nu * (Ve - Vi)), "\n")
print("  mi_me*Grad_par(phi)", exprmag(mi_me * Grad_par(phi, metric)), "\n")
print(
    "  mi_me*Te*Grad_par(log(Ne))",
    exprmag(-mi_me * Te * Grad_par(log(Ne), metric)),
    "\n",
)
print(
    "  mi_me*1.71*Grad_par(Te) : ", exprmag(-mi_me * 1.71 * Grad_par(Te, metric)), "\n"
)


print("\nVe mag: ", exprmag(Ve), exprmag(VePsi), exprmag(0.5 * mi_me * beta_e * psi))

print("\n\n########################################")
print("\n\nVi terms:")

print("  bracket : ", exprmag(-bracket(phi, Vi, metric)), "\n")
print("  Vpar_Grad_par(Vi, Vi)", exprmag(-Vpar_Grad_par(Vi, Vi, metric)), "\n")
print(
    "  (2./3.)*Grad_par(Gi)", exprmag((old_div(2.0, 3.0)) * Grad_par(Gi, metric)), "\n"
)
print("  Grad_par(Te", exprmag(-Grad_par(Te, metric)), "\n")
print("  Te*Grad_par(log(Ne))", exprmag(-Te * Grad_par(log(Ne), metric)), "\n")
